import sqlite3
import yaml
from typing import Optional, Union
from contextlib import closing

import numpy as np
import pandas as pd
import pysam
from jaxtyping import Float

from nanocompore.common import MEASUREMENTS_TYPE
from nanocompore.common import INTENSITY_POS
from nanocompore.common import DWELL_POS
from nanocompore.common import UNCALLED4
from nanocompore.common import Kit
from nanocompore.common import get_references_from_bam
from nanocompore.config import Config
from nanocompore.database import PreprocessingDB
from nanocompore.uncalled4 import Uncalled4


def load_config(config_path: str) -> Config:
    """
    Load a configuration file.

    Parameters
    ----------
    config_path : str
        Path to the Nanocompore configuration file.

    Returns
    -------
    Config
        A configuration object.
    """
    with open(config_path, 'rb') as f:
        return Config(yaml.safe_load(f))


def get_references(config: Config, has_data=True) -> list[str]:
    """
    Returns a list of all references found in the
    list of samples defined in the configuration.

    Parameters
    ----------
    config : Config
        Path to a Nanocompore configuration file.
    has_data : bool, default=True
        If True (default) will return only references
        for which there are mapped reads.

    Returns
    -------
    list
        List of transcript reference id strings.

    Examples
    --------
    >>> from nanocompore.api import load_config, get_references
    >>> config = load_config('analysis.yaml')
    >>> get_references(config)
    ['ENST00000674681.1|ENSG00000075624.17|OTTHUMG00000023268|-|ACTB-219|ACTB|2554|protein_coding|', 'ENST00000642480.2|ENSG00000075624.17|OTTHUMG00000023268|OTTHUMT00000495153.1|ACTB-213|ACTB|2021|protein_coding|']
    """
    data_files = list(_get_data_files(config).values())
    if config.get_resquiggler() == UNCALLED4:
        return _get_bam_references(data_files, has_data)
    else:
        return _get_db_references(data_files, has_data)


def get_reads(
        config: Config,
        reference_id: str,
        selected_reads: Optional[list[str]]=None
    ) -> tuple[Float[np.ndarray, "reads positions variables"],
               list[str],
               list[str],
               list[str]]:
    """
    Get the data for all reads mapping to the given reference.

    Parameters
    ----------
    config : Config
        Path to a Nanocompore configuration file.
    reference_id : str
        ID for a reference sequence (transcript).
    selected_reads : Optional[list[str]]
        Optional list of UUIDs of the reads for which to get data.
        By default it's set to None and returns all reads.

    Returns
    -------
    tuple[Float[np.ndarray, ["reads positions variables"]],
          list[str],
          list[str],
          list[str]]

        A tuple with (signal_data, reads, samples, conditions)

        - signal_data is a 3D array with shape (reads, positions, variables).
          In the variables dimension 0=intensity, 1=dwell time.
        - reads is a list of read ids (qname).
        - samples is a list of the sample labels (as defined in the config).
        - conditions is a list of the condition labels (as defined in the config).

    Raises
    ------
    KeyError
        If the reference_id is not found in the data sources.

    Examples
    --------
    >>> from nanocompore.api import load_config, get_references
    >>> config = load_config('analysis.yaml')
    >>> get_reads(config, 'ENST00000674681.1|ENSG00000075624.17|OTTHUMG00000023268|-|ACTB-219|ACTB|2554|protein_coding|')
    """
    data_files = _get_data_files(config)
    if config.get_resquiggler() == UNCALLED4:
        kit = config.get_kit()
        data, reads, samples = _get_bam_reads(data_files.values(),
                                              reference_id,
                                              kit,
                                              selected_reads)
    else:
        data, reads, samples = _get_db_reads(data_files.values(),
                                             reference_id,
                                             selected_reads)
    sample_mapper = np.vectorize(dict(enumerate(data_files)).get)
    samples = sample_mapper(samples)
    condition_mapper = np.vectorize(config.sample_to_condition().get)
    conditions = condition_mapper(samples)
    return data, reads, samples.tolist(), conditions.tolist()


def get_pos(config: Config, reference_id: str, pos: int) -> pd.DataFrame:
    """
    Get the data for a given position for all samples.
    Note that position is a 0-based index of the first
    nucleotide of a k-mer.

    Returns the signal data for a specific position
    of the given reference transcript from all reads.

    Parameters
    ----------
    config : Config
        Path to a Nanocompore configuration file.
    reference_id : str
        ID for a reference sequence (transcript).
    pos : int
        Position on the transcript for which to get data. A 0-based
        index is assumed.

    Returns
    -------
    pandas.DataFrame
        Where the DataFrame contains the following columns:

        - condition  condition label (as defined in the configuration)
        - sample     sample label (as defined in the configuration)
        - read:      id of the read (qname)
        - intensity: current intensity
        - dwell:     dwell time for the kmer

    Examples
    --------
    >>> from nanocompore.api import load_config, get_pos
    >>> config = load_config('analysis.yaml')
    >>> get_pos(config, 'ENST00000674681.1|ENSG00000075624.17|OTTHUMG00000023268|-|ACTB-219|ACTB|2554|protein_coding|', 532)
       condition sample                                  read  intensity  dwell
    0         WT    WT1  a4395b0d-dd3b-48e3-8afb-4085374b1147     3800.0    7.0
    1         WT    WT1  f9733448-6e6b-47ba-9501-01eda2f5ea26     4865.0  126.0
    2         WT    WT1  6f5e3b2e-f27b-47ef-b3c6-2ab4fdefd20a     3272.0   42.0
    3         WT    WT2  2da07406-70c2-40a1-835a-6a7a2c914d49     6241.0   44.0
    4         WT    WT2  54fc1d38-5e3d-4d77-a717-2d41b4785af6     4047.0    9.0
    5         WT    WT2  3cfa90d1-7dfb-4398-a224-c75a3ab99873     3709.0   70.0
    6         KD    KD1  3f46f499-8ce4-4817-8177-8ad61b784f27     4807.0   57.0
    7         KD    KD1  73d62df4-f04a-4207-a4bc-7b9739b3c3b2     4336.0  132.0
    8         KD    KD1  b7bc9a36-318e-4be2-a90f-74a5aa6439bf     -861.0    7.0
    9         KD    KD2  ac486e16-15be-47a8-902c-2cfa2887c534     2706.0   45.0
    10        KD    KD2  797fd991-570e-42d4-8292-0a7557b192d7     5450.0   24.0
    11        KD    KD2  4e1ad358-ec2b-40b4-8e9a-54db28a40551      206.0   47.0
    """
    data_files = _get_data_files(config)
    sample_mapper = np.vectorize(dict(enumerate(data_files)).get)
    if config.get_resquiggler() == UNCALLED4:
        kit = config.get_kit()
        df = _get_bam_pos(data_files.values(), reference_id, pos, kit)
    else:
        df = _get_db_pos(data_files.values(), reference_id, pos)
    df['sample'] = sample_mapper(df['sample'])
    condition_mapper = np.vectorize(config.sample_to_condition().get)
    df['condition'] = condition_mapper(df['sample'])
    return df.loc[:, ['condition', 'sample', 'read', 'intensity', 'dwell']]


def _get_bam_references(bams: Union[str, list[str]], has_data=True) -> list[str]:
    """
    Returns a list of all references in the given
    BAM.

    Parameters
    ----------
    bams : Union[str, list[str]]
        One on more paths to BAM files produced by Uncalled4.
    has_data : bool, default=True
        If True (default) will return only references
        for which there are mapped reads.

    Returns
    -------
    list
        List of transcript reference id strings.

    Examples
    --------
    >>> from nanocompore.api import get_bam_references
    >>> get_bam_references('wt1.bam')
    ['ENST00000674681.1|ENSG00000075624.17|OTTHUMG00000023268|-|ACTB-219|ACTB|2554|protein_coding|', 'ENST00000642480.2|ENSG00000075624.17|OTTHUMG00000023268|OTTHUMT00000495153.1|ACTB-213|ACTB|2021|protein_coding|']
    """
    if isinstance(bams, str):
        return list(get_references_from_bam(bams, has_data).keys())
    return list({ref
                 for bam in bams
                 for ref in get_references_from_bam(bam, has_data).keys()})


def _get_db_references(dbs: Union[str, list[str]], has_data=True) -> list[str]:
    """
    Returns a list of all references in the given
    databases.

    Parameters
    ----------
    dbs : Union[str, list[str]]
        One or more paths to SQLite databases produced by
        Nanocompore's eventalign_collapse or remora_resquiggle
        subcommands.
    has_data : bool, default=True
        If True (default) will return only
        references for which there are data.

    Returns
    -------
    list
        List of transcript reference id strings.

    Examples
    --------
    >>> from nanocompore.api import get_db_references
    >>> get_db_references('wt1.sqlite')
    ['ENST00000674681.1|ENSG00000075624.17|OTTHUMG00000023268|-|ACTB-219|ACTB|2554|protein_coding|', 'ENST00000642480.2|ENSG00000075624.17|OTTHUMG00000023268|OTTHUMT00000495153.1|ACTB-213|ACTB|2021|protein_coding|']
    """
    references = set()
    for db in dbs:
        if has_data:
            transcripts = PreprocessingDB(db).get_references_with_data()
            references.update(transcripts.keys())
        else:
            transcripts = PreprocessingDB(db).get_references()
            references.update([t.ref_id for t in transcripts])
    return list(references)


def _get_bam_reads(
        bams: Union[str, list[str]],
        reference_id: str,
        kit: Kit,
        selected_reads: Optional[list[str]]=None
    ) -> tuple[Float[np.ndarray, "reads positions variables"],
               list[str],
               list[str]]:
    """
    Get a numpy array with read data for specific reference_id from
    one or more BAMs produced by Uncalled4.

    Parameters
    ----------
    bams : str
        Paths to one or more BAM files.
    reference_id : str
        ID for a reference sequence (transcript).
    kit : Kit
        The chemistry used for the sequencing (e.g. RNA002 or RNA004).
    selected_reads : Optional[list[str]]
        Optional list of UUIDs of the reads for which to get data.
        By default it's set to None and returns all reads.

    Returns
    -------
    tuple[Float[np.ndarray, ["reads positions variables"]],
          list[str]]

        A tuple with (signal_data, reads, samples)

        - signal_data is a 3D array with shape (reads, positions, variables).
          In the variables dimension 0=intensity, 1=dwell time.
        - reads is a list of read ids (qname).
        - samples is a list of the sample ids (in the
          order in which the bams were provided).

    Raises
    ------
    KeyError
        If the reference_id is not found in the BAMs.

    Examples
    --------
    >>> from nanocompore.api import get_bam_reads, Kit
    >>> get_bam_reads(['path/to/wt.bam', 'path/to/kd.bam'], 'ENST00000674681.1|ENSG00000075624.17|OTTHUMG00000023268|-|ACTB-219|ACTB|2554|protein_coding|', Kit.RNA004)
    """
    if isinstance(bams, str):
        bams = [bams]
    opened_bams = {i: pysam.AlignmentFile(bam, 'rb')
                   for i, bam in enumerate(bams)}
    ref_len = None
    for bam in opened_bams.values():
        try:
            ref_len = bam.get_reference_length(reference_id)
            break
        except KeyError:
            pass
    else:
        raise KeyError(f"Reference {reference_id} not found in the bams.")

    uncalled4 = Uncalled4(reference_id,
                          ref_len,
                          opened_bams,
                          kit)
    data, reads, samples = uncalled4.get_data()
    if selected_reads:
        read_set = set(selected_reads)
        mask = np.array([read in read_set for read in reads])
        return data[mask], reads[mask].tolist(), samples[mask].tolist()
    return np.moveaxis(data, (0, 1, 2), (1, 0, 2)), reads.tolist(), samples.tolist()


def _get_db_reads(
        dbs: Union[str, list[str]],
        reference_id: str,
        selected_reads: Optional[list[str]]=None
    ) -> tuple[
            Float[np.ndarray, "reads positions variables"],
            list[str],
            list[str]]:
    """
    Get a numpy array with read data for specific reference_id from
    one or more preprocessing databases produced by eventalign_collapse or
    remora_resquiggle.

    Parameters
    ----------
    dbs : Union[str, list[str]]
        Paths to the SQLite database files.
    reference_id : str
        ID for a reference sequence (transcript).
    selected_reads : Optional[list[str]]
        Optional list of UUIDs of the reads for which to get data.
        By default it's set to None and returns all reads.

    Returns
    -------
    tuple[Float[np.ndarray, ["reads positions variables"]],
          list[str]]

        A tuple with (signal_data, reads, samples)

        - signal_data is a 3D array with shape (reads, positions, variables).
          In the variables dimension 0=intensity, 1=dwell time.
        - reads is a list of read ids (qname).
        - samples is a list of the sample ids (in the
          order in which the dbs were provided).

    Examples
    --------
    >>> from nanocompore.api import get_db_reads
    >>> get_db_reads(['path/to/wt.sqlite', 'path/to/kd.sqlite'], 'ENST00000674681.1|ENSG00000075624.17|OTTHUMG00000023268|-|ACTB-219|ACTB|2554|protein_coding|')
    """
    all_data = []
    all_reads = []
    all_samples = []
    for i, db in enumerate(dbs):
        data, reads = _get_sample_db_reads(db, reference_id, selected_reads)
        if len(reads) > 0:
            all_data.append(data)
            all_reads.extend(reads)
            all_samples.extend(np.full((len(reads),), i))
    return np.concatenate(all_data), all_reads, all_samples


def _get_sample_db_reads(
        db: str,
        reference_id: str,
        selected_reads: Optional[list[str]]=None
    ) -> tuple[
            Union[Float[np.ndarray, "reads positions variables"], None],
            list[str]]:
    """
    Get a numpy array with read data for specific reference_id from
    a preprocessing database produced by eventalign_collapse or
    remora_resquiggle.

    Parameters
    ----------
    db : str
        Path to the SQLite database file.
    reference_id : str
        ID for a reference sequence (transcript).
    selected_reads : Optional[list[str]]
        Optional list of UUIDs of the reads for which to get data.

    Returns
    -------
    tuple[Union[Float[np.ndarray, ["reads positions variables"]], None],
          list[str]]

        A tuple with (signal_data, reads)

        - signal_data is a 3D array with shape (reads, positions, variables).
          In the variables dimension 0=intensity, 1=dwell time. If no reads
          are found for the reference, signal_data will be None.
        - reads is a list of read ids (bam qname).

    Examples
    --------
    >>> from nanocompore.api import get_sample_db_reads
    >>> get_sample_db_reads('path/to/wt.sqlite', 'ENST00000674681.1|ENSG00000075624.17|OTTHUMG00000023268|-|ACTB-219|ACTB|2554|protein_coding|')
    """
    with closing(sqlite3.connect(db)) as conn,\
         closing(conn.cursor()) as cursor:
        rows = cursor.execute("""
            SELECT intensity, dwell, read
            FROM signal_data sd
            JOIN transcripts t ON t.id = sd.transcript_id
            JOIN reads r ON r.id = sd.read_id
            WHERE t.name = ?""", (reference_id,)).fetchall()
        intensity = np.array([np.frombuffer(row[0], dtype=MEASUREMENTS_TYPE)
                              for row in rows])
        dwell = np.array([np.frombuffer(row[1], dtype=MEASUREMENTS_TYPE)
                          for row in rows])
        reads = [row[2] for row in rows]
        if len(reads) == 0:
            return None, reads
        signal_data = np.dstack([intensity, dwell])
        if selected_reads:
            read_set = set(selected_reads)
            mask = np.array([read in read_set for read in reads])
            filtered_reads = [read for read in reads if read in read_set]
            return signal_data[mask], filtered_reads
        return signal_data, reads


def _get_db_read(db: str, read_id: str) -> Float[np.ndarray, "reads positions variables"]:
    """
    Get a numpy array with the signal data for a specific read from
    a preprocessing database produced by eventalign_collapse or
    remora_resquiggle.

    Parameters
    ----------
    db : str
        Path to the SQLite database file.
    read_id : str
        ID of the read for which to get data.

    Returns
    -------
    Float[np.ndarray, "positions variables"]
        Sinal data is a 2D numpy array with shape
        (positions, variables). Variables
        will be 0=intensity and 1=dwell time.

    Examples
    --------
    >>> from nanocompore.api import get_db_read
    >>> get_db_read('path/to/wt.sqlite', 'f9733448-6e6b-47ba-9501-01eda2f5ea26')
    """
    with closing(sqlite3.connect(db)) as conn,\
         closing(conn.cursor()) as cursor:
        row = cursor.execute("""
            SELECT intensity, dwell
            FROM signal_data sd
            JOIN reads r ON r.id = sd.read_id
            WHERE r.read = ?
            LIMIT 1""", (read_id,)).fetchone()
        intensity = np.frombuffer(row[0], dtype=MEASUREMENTS_TYPE)
        dwell = np.frombuffer(row[1], dtype=MEASUREMENTS_TYPE)
        return np.stack([intensity, dwell], axis=1)


def _get_bam_pos(bams: Union[str, list[str]], reference_id: str, pos: int, kit: Kit) -> pd.DataFrame:
    """
    Get data for a position from one or more BAMs created by Uncalled4.

    Returns the signal data for a specific position
    of the given reference transcript from all reads.

    Parameters
    ----------
    bams : Union[str, list[str]]
        Paths to the BAM files produced by Uncalled4.
    reference_id : str
        ID for a reference sequence (transcript).
    pos : int
        Position on the transcript for which to get data. A 0-based
        index is assumed.
    kit : Kit
        The chemistry used for the sequencing (e.g. RNA002 or RNA004).


    Returns
    -------
    pandas.DataFrame
        Where the DataFrame contains the following columns:

        - sample     id of the sample (equal to the index
                     of the bam in the bams parameter).
        - read:      id of the read
        - intensity: current intensity
        - dwell:     dwell time for the kmer

    Examples
    --------
    >>> from nanocompore.api import get_bam_pos
    >>> get_bam_pos(['wt1.bam', 'kd1.bam'], 'ENST00000674681.1|ENSG00000075624.17|OTTHUMG00000023268|-|ACTB-219|ACTB|2554|protein_coding|', 532, Kit.RNA004)
       sample                                  read  intensity  dwell
    0       0  a4395b0d-dd3b-48e3-8afb-4085374b1147     3800.0    7.0
    1       0  f9733448-6e6b-47ba-9501-01eda2f5ea26     4865.0  126.0
    2       0  6f5e3b2e-f27b-47ef-b3c6-2ab4fdefd20a     3272.0   42.0
    3       1  3f46f499-8ce4-4817-8177-8ad61b784f27     4807.0   57.0
    4       1  73d62df4-f04a-4207-a4bc-7b9739b3c3b2     4336.0  132.0
    5       1  b7bc9a36-318e-4be2-a90f-74a5aa6439bf     -861.0    7.0
    """
    reads_data, reads, samples = _get_bam_reads(bams, reference_id, kit)
    pos_data = reads_data[:, pos, :]
    valid = ~np.isnan(pos_data[:, 0])
    return pd.DataFrame({'sample': np.array(samples)[valid],
                         'read': np.array(reads)[valid],
                         'intensity': pos_data[valid, INTENSITY_POS],
                         'dwell': pos_data[valid, DWELL_POS]})


def _get_db_pos(dbs: Union[str, list[str]], reference_id: str, pos: int) -> pd.DataFrame:
    """
    Get data for a position for a single sample from SQLite databases
    created by the eventalign_collapse or remora_resquiggle subcommands.

    Returns the signal data for a specific position
    of the given reference transcript from all reads.

    Parameters
    ----------
    dbs : Union[str, list[str]]
        Paths to the SQLite databases produced by eventalign_collapse
        or remora_resquiggle.
    reference_id : str
        ID for a reference sequence (transcript).
    pos : int
        Position on the transcript for which to get data. A 0-based
        index is assumed.

    Returns
    -------
    pandas.DataFrame
        Where the DataFrame contains the following columns:

        - sample     id of the sample (equal to the index
                     of the db in the dbs parameter).
        - read:      id of the read
        - intensity: current intensity
        - dwell:     dwell time for the kmer
    """
    reads_data, reads, samples = _get_db_reads(dbs, reference_id)
    pos_data = reads_data[:, pos, :]
    valid = ~np.isnan(pos_data[:, 0])
    return pd.DataFrame({'sample': np.array(samples)[valid],
                         'read': np.array(reads)[valid],
                         'intensity': pos_data[valid, INTENSITY_POS],
                         'dwell': pos_data[valid, DWELL_POS]})


def get_metadata(db: str) -> dict[str, str]:
    """
    Returns the metadata from the given SQLite database.

    The metadata contains information such as input files,
    resquiggler used, and data types for the binary encoded fields.

    Parameters
    ----------
    db : str
        Path to the SQLite database produced by the preprocessing
        command of Nanocompore.

    Returns
    -------
    dict
        Dictionary containing the metadata
    """
    with closing(sqlite3.connect(db)) as conn,\
         closing(conn.cursor()) as cursor:
        query = "SELECT key, value FROM metadata"
        return {k: v for k, v in cursor.execute(query).fetchall()}


def _get_data_files(config: Config) -> dict[str, str]:
    field = 'db'
    if config.get_resquiggler() == UNCALLED4:
        field = 'bam'
    return {sample: sample_data[field]
            for cond_data in config.get_data().values()
            for sample, sample_data in cond_data.items()}




import gc
import time
from collections import defaultdict
from typing import Union

import numpy as np
import pandas as pd
import torch
from gmm_gpu.gmm import GMM
from jaxtyping import Float, Int
from loguru import logger
from numpy.lib.stride_tricks import sliding_window_view
from scipy.stats import mannwhitneyu, ttest_ind, chi2_contingency, chi2
from scipy.stats.mstats import ks_twosamp

from nanocompore.common import NanocomporeError
from nanocompore.gof_tests import gof_test_multirep
from nanocompore.gof_tests import gof_test_singlerep
from nanocompore.transcript import Transcript


INTENSITY = 0
DWELL = 1
MOTOR = 2


class TranscriptComparator:
    """
    Compare a set of positions.
    """
    def __init__(self, config, worker, random_seed=42, dtype=torch.float32):
        self._config = config
        self._random_seed = random_seed
        self._dtype = dtype
        self._motor_dwell_offset = self._config.get_motor_dwell_offset()
        self._worker = worker


    def compare_transcript(
        self,
        transcript: Transcript,
        data: Float[torch.Tensor, "positions reads vars"],
        samples: Int[torch.Tensor, "reads"],
        conditions: Int[torch.Tensor, "reads"],
        positions: Int[torch.Tensor, "positions"],
        device: str
    ) -> tuple[Transcript, Union[pd.DataFrame, None]]:
        """
        Compare the two conditions for the given transcript.

        Parameters
        ----------
        transcript : Transcript
            The transcript for which the two conditions
            will be compared.
        data : Float[torch.Tensor, "positions reads vars"]
            Torch tensor with shape (positions, reads, vars)
            containing the data for the comparison.
        samples : Int[torch.Tensor, "reads"]
            Array of sample ids with size *reads*.
        conditions : Int[torch.Tensor, "reads"]
            Array of condition ids with size *reads*.
        positions : Int[torch.Tensor, "positions"]
            Array of reference positions on the transcript
            with size *positions*.
        device : str
            String indicating the device that will be used
            for the comparison. Indicates if CPU or GPU
            should be used.

        Returns
        -------
        tuple[Transcript, pd.DataFrame]
            Tuple with the Transcript and the comparison
            results stored in a pandas DataFrame.
        """
        n_positions = len(positions)
        if n_positions == 0:
            return (transcript, None)

        results = pd.DataFrame({'transcript_id': transcript.id,
                                'pos': positions.cpu()})

        self._worker.log("trace", "Start shift stats")
        t = time.time()
        self.retry(lambda: self._add_shift_stats(results, data, conditions, device),
                   exception=torch.OutOfMemoryError)
        if device.startswith('cuda'):
            torch.cuda.empty_cache()
            torch.cuda.reset_peak_memory_stats()
        self._worker.log("debug", f"Finished shift stats ({time.time() - t})")

        std = nanstd(data, 1).unsqueeze(1)
        outliers = (((data - data.nanmean(1, keepdim=True)) / std).abs() > 3).any(2)
        data[outliers] = np.nan

        # Standardize the data
        std = nanstd(data, 1)
        data = (data - data.nanmean(1).unsqueeze(1)) / std.unsqueeze(1)

        # If we get the same value for a variable in most of the reads
        # the others will be removed as outliers and we cannot calculate
        # the std to scale the data. I've noticed this only in Uncalled4
        # because sometimes it would return the -32768 for the intensity
        # for all/most reads.
        bad_stds = std.isnan().any(1) | (std == 0).any(1)
        if bad_stds.any():
            self._worker.log(
                    "warning",
                    "The standard deviation cannot be calculated for some positions on "
                    f"transcript {transcript.name}, but are required for scaling the data. "
                    f"The positions {positions[bad_stds].tolist()} will be skipped.")
            data = data[~bad_stds]
            positions = positions[~bad_stds]
            n_positions = positions.shape[0]
            results.drop(np.arange(results.shape[0])[bad_stds.cpu()],
                         inplace=True,
                         axis=0)

        auto_test_mask = None
        has_auto = 'auto' in self._config.get_comparison_methods()
        if has_auto:
            auto_test_mask = self._auto_test_mask(data)
        test_masks = self._get_test_masks(auto_test_mask, n_positions)

        for test, mask in test_masks.items():
            self._worker.log("debug", f"Start {test}")
            t = time.time()
            test_results = self._run_test(test,
                                          data[mask, :, :],
                                          samples,
                                          conditions,
                                          device=device)
            self._worker.log("debug", f"Finished {test} ({time.time() - t})")
            self._merge_results(results, test_results, test, mask, auto_test_mask, n_positions)

        context = self._config.get_sequence_context()
        if context > 0:
            tests = [col for col in results.columns if 'pvalue' in col]
            for test in tests:
                combined_pvals = self._combine_context_pvalues(results, test)

                label = f"{test}_context_{context}"
                results[label] = combined_pvals

        return transcript, results


    def _get_test_masks(self, auto_test_mask, n_positions):
        base_tests = set(self._config.get_comparison_methods()) - {'auto'}
        if auto_test_mask is not None:
            additional_tests = set(auto_test_mask) - base_tests
        else:
            additional_tests = []
        masks = {}
        for test in base_tests:
            masks[test] = np.ones(n_positions, dtype=bool)
        for test in additional_tests:
            masks[test] = auto_test_mask == test
        return masks


    def _auto_test_mask(self, data):
        cov = (~data[:, :, INTENSITY].isnan()).cpu().sum(1).numpy()
        return np.where(cov < 500, 'KS', 'GMM')


    def _auto_test_pvalue(self, test):
        if test == 'KS':
            return 'KS_intensity_pvalue'
        elif test == 'GMM':
            return 'GMM_chi2_pvalue'
        elif test == 'GOF':
            return 'GOF_pvalue'
        else:
            raise NotImplementedError("Unhandled test {test} as automatic test.")


    def _transcript_reads(self, kmers):
        """
        Get a list of kmers and returns a dict of type:
        transcript_id -> {read_id -> index}
        """
        transcript_reads = defaultdict(set)
        for kmer in kmers:
            transcript_reads[kmer.transcript_id].update(kmer.reads)
        return {ref_id: dict(zip(reads, range(len(reads))))
                for ref_id, reads in transcript_reads.items()}


    def _run_test(self, test, test_data, samples, conditions, device):
        if test.upper() in ['MW', 'KS', 'TT']:
            return self._nonparametric_test(test, test_data, conditions)
        elif test.upper() == 'GMM':
            return self._gmm_test(test_data, samples, conditions, device)
        elif test.upper() == 'GOF':
            return self._gof_test(test_data, samples, conditions)
        else:
            raise NotImplementedError(f"Test {test} is not implemented!")


    def _merge_results(self, results, test_results, test, mask, auto_test_mask, n_positions):
        has_auto = 'auto' in self._config.get_comparison_methods()

        if has_auto and 'auto_pvalue' not in results:
            results['auto_pvalue'] = np.full(n_positions, np.nan)
            results['auto_test'] = auto_test_mask

        for column, values in test_results.items():
            results.loc[mask, column] = values
            if has_auto and column == self._auto_test_pvalue(test):
                rows = auto_test_mask == test
                results.loc[rows, 'auto_pvalue'] = results.loc[rows, column]


    def _nonparametric_test(self, test, test_data, conditions):
        if test in ["mann_whitney", "MW"]:
            stat_test = lambda x, y: mannwhitneyu(x, y, alternative='two-sided')
        elif test in ["kolmogorov_smirnov", "KS"]:
            stat_test = ks_twosamp
        elif test in ["t_test", "TT"]:
            stat_test = lambda x, y: ttest_ind(x, y, equal_var=False)
        else:
            raise NanocomporeError("Invalid statistical method name (MW, KS, TT)")

        cond0_mask = conditions == 0

        intensity_pvals = []
        dwell_pvals = []
        for i in range(test_data.shape[0]):
            cond0_data = test_data[i, cond0_mask, :]
            cond1_data = test_data[i, ~cond0_mask, :]
            try:
                pval = stat_test(self._drop_nans(cond0_data[:, INTENSITY]).cpu(),
                                 self._drop_nans(cond1_data[:, INTENSITY]).cpu()).pvalue
                intensity_pvals.append(pval)
            except Exception as e:
                self._worker.log("error",
                                 f"Error in nonparametric test on the intensity: {e}")
                intensity_pvals.append(np.nan)
            try:
                pval = stat_test(self._drop_nans(cond0_data[:, DWELL]).cpu(),
                                 self._drop_nans(cond1_data[:, DWELL]).cpu()).pvalue
                dwell_pvals.append(pval)
            except Exception as e:
                self._worker.log("error",
                                 f"Error in nonparametric test on the dwell time: {e}")
                dwell_pvals.append(np.nan)
        return {f'{test}_intensity_pvalue': intensity_pvals,
                f'{test}_dwell_pvalue': dwell_pvals}


    def _drop_nans(self, values):
        return values[~values.isnan()]


    def _gof_test(self, test_data, samples, conditions):
        if self._config.is_multi_replicate():
            return self._gof_test_multirep(test_data, samples, conditions)
        else:
            return self._gof_test_singlerep(test_data, conditions)


    def _gof_test_singlerep(self, test_data, conditions):
        pvals = []
        for i in range(test_data.shape[0]):
            pval = gof_test_singlerep(test_data[i, :, :], conditions, self._config)
            pvals.append(pval)
        return {'GOF_pvalue': pvals}


    def _gof_test_multirep(self, test_data, samples, conditions):
        pvals = []
        for i in range(test_data.shape[0]):
            pval = gof_test_multirep(test_data[i, :, :],
                                     samples,
                                     conditions,
                                     self._config)
            pvals.append(pval)
        return {'GOF_pvalue': pvals}


    def _gmm_test(self, test_data, samples, conditions, device):
        # If we don't use the motor dwell time
        # we would get a tensor that only contains
        # intensity and dwell and we can directly
        # run the tests.
        if self._motor_dwell_offset == 0:
            return self._gmm_test_split(test_data, samples, conditions, device)

        # If we use the motor dwell time than we have
        # to consider some additional complications:
        # For each position some reads have motor dwell
        # and others don't. We split the positions into
        # two sets: one set of positions where all reads
        # have motor dwell time - those would be tested
        # with a 3D GMM - and a second set of positions
        # where some reads are missing the motor dwell -
        # those would be tested with a 2D GMM.
        dim3_data, dim2_data, split = self._split_by_ndim(test_data)
        columns = set()
        if dim3_data.shape[0] > 0:
            dim3_results = self._gmm_test_split(dim3_data,
                                                samples,
                                                conditions,
                                                device)
            columns.update(dim3_results.keys())
        if dim2_data.shape[0] > 0:
            dim2_results = self._gmm_test_split(dim2_data,
                                                samples,
                                                conditions,
                                                device)
            columns.update(dim2_results.keys())

        results = {}
        for column in columns:
            if dim3_data.shape[0] > 0:
                dim3_values = dim3_results[column]
            if dim2_data.shape[0] > 0:
                dim2_values = dim2_results[column]
            dim3_i = 0
            dim2_i = 0

            merged = []
            for s in split:
                if s == 0:
                    merged.append(dim3_values[dim3_i])
                    dim3_i += 1
                else:
                    merged.append(dim2_values[dim2_i])
                    dim2_i += 1
            results[column] = merged
        return results


    def _gmm_test_split(self, data, samples, conditions, device):
        test_data = data.to(self._dtype)
        s = time.time()
        def fit_model(components):
            gmm = GMM(n_components=components,
                      device=device,
                      random_seed=self._random_seed,
                      dtype=self._dtype)
            gmm.fit(test_data)
            return gmm
        gmm1 = self.retry(lambda: fit_model(1), exception=torch.OutOfMemoryError)
        bic1 = gmm1.bic(test_data)
        del gmm1
        gc.collect()
        gmm2 = self.retry(lambda: fit_model(2), exception=torch.OutOfMemoryError)

        self._worker.log("debug", f"GMM fitting time: {time.time() - s}")
        pred = gmm2.predict(test_data, force_cpu_result=False)
        pred = torch.where(test_data[:, :, 0].isnan(), np.nan, pred)

        bic2 = gmm2.bic(test_data)
        del gmm2
        gc.collect()
        if device.startswith('cuda'):
            torch.cuda.empty_cache()
            torch.cuda.reset_peak_memory_stats()

        contingencies = get_contigency_matrices(conditions, pred).to(device=device)
        counts = self._get_cluster_counts(contingencies, samples, conditions, pred)
        # We add 1 to all cells in all contingency
        # matrices to make sure we don't encounter
        # a devision by zero.
        contingencies += 1
        lors = calculate_lors(contingencies)
        pvals = np.array([chi2_contingency(cont).pvalue
                          for cont in contingencies.cpu().numpy()])
        # If a single component fits the data
        # better we ignore the results from
        # the two-component GMM.
        ignored_tests = (bic1 <= bic2).numpy()
        lors[ignored_tests] = np.nan
        pvals[ignored_tests] = np.nan

        return {'GMM_chi2_pvalue': pvals,
                'GMM_LOR': lors.cpu().numpy()} | counts


    def _split_by_ndim(self, test_data):
        valid = ~test_data.isnan()
        with_motor = (valid[:, :, INTENSITY] & valid[:, :, MOTOR]).sum(1)
        total = valid[:, :, INTENSITY].sum(1)
        use_motor = with_motor >= 0.9 * total

        return (test_data[use_motor],
                test_data[~use_motor, :, :MOTOR],
                (~use_motor).int())


    def _add_shift_stats(self, results, data, conditions, device):
        data = data.to(device)

        stats = {0: {}, 1: {}}
        cond0_mask = (conditions == 0)[None, :, None].to(device)

        cond0_data = torch.where(cond0_mask, data, np.nan)
        stats[0]['mean'] = cond0_data.nanmean(1).cpu()
        stats[0]['median'] = cond0_data.nanmedian(1).values.cpu()
        stats[0]['sd'] = nanstd(cond0_data, 1).cpu()

        cond1_data = torch.where(~cond0_mask, data, np.nan)
        stats[1]['mean'] = cond1_data.nanmean(1).cpu()
        stats[1]['median'] = cond1_data.nanmedian(1).values.cpu()
        stats[1]['sd'] = nanstd(cond1_data, 1).cpu()

        dims = {
                'intensity': INTENSITY,
                'dwell': DWELL,
               }
        if self._motor_dwell_offset > 0:
            dims['motor_dwell'] = MOTOR

        for cond in [0, 1]:
            for stat in ['mean', 'median', 'sd']:
                for dim, dim_index in dims.items():
                    label = f"c{cond+1}_{stat}_{dim}"
                    results[label] = stats[cond][stat][:, dim_index]


    def _combine_context_pvalues(self, results, test):
        sequence_context = self._config.get_sequence_context()
        sequence_context_weights = self._config.get_sequence_context_weights()
        if sequence_context_weights == "harmonic":
            # Generate weights as a symmetrical harmonic series
            weights = harmomic_series(sequence_context)
        else:
            weights = [1]*(2*sequence_context + 1)

        min_pos = results.pos.min()
        max_pos = results.pos.max()

        # We want to combine pvalues of neighbouring
        # positions, but the results dataframe may
        # have missing positions.
        # Additionally, applying a window function
        # means we lose the results for the positions
        # at the ends so we have to pad the results.
        padded_size = max_pos + 1 - min_pos + 2*sequence_context
        pvalues = np.ones(padded_size, dtype=float)
        indices = results.pos - min_pos
        pvalues[indices + sequence_context] = results[test]

        corr = cross_corr_matrix(pvalues[sequence_context:-sequence_context],
                                 sequence_context)

        window_size = 2*sequence_context + 1
        def func(window):
            if np.isnan(window[sequence_context]):
                return np.nan
            else:
                return combine_pvalues_hou(np.nan_to_num(window, nan=1),
                                           weights,
                                           corr)
        combinator = np.vectorize(func, signature='(n)->()')
        windows = sliding_window_view(pvalues, window_shape=window_size)
        combined_pvals = np.apply_along_axis(combinator, 1, windows)

        return combined_pvals[indices]


    def _get_cluster_counts(self, contingency, samples, conditions, predictions):
        depleted_cond = self._config.get_depleted_condition()
        depleted_cond_id = self._config.get_condition_ids()[depleted_cond]

        cond_counts = contingency.sum(2).to(torch.uint32)
        # contingency will be something like this:
        # e.g.      cluster
        # condition   0   1
        #         0  100  30
        #         1   90  40
        freq_contingency = contingency / cond_counts.unsqueeze(2)

        # We assume that the unmodified cluster is the one
        # in which the depleted condition has most of its
        # points. Hence, the modified is the one, where
        # the depleted condition has fewer points.
        # I.e. if the condition 0 is the depleted one,
        # in the example contingency above the modified
        # cluster will be cluster 1.
        mod_clusters = freq_contingency[:, depleted_cond_id, :].argmin(1)
        mod_clusters = mod_clusters.unsqueeze(1).expand_as(predictions)

        # Iterate all samples and calculate the
        # number of modified and non-modified reads.
        cluster_counts = {}
        for sample_label, sample in self._config.get_sample_ids().items():
            sample_predictions = predictions[:, samples == sample]
            sample_mod_clusters = mod_clusters[:, samples == sample]
            sample_totals = torch.sum(~sample_predictions.isnan(), dim=1)
            mod_counts = (sample_predictions == sample_mod_clusters).sum(dim=1)
            unmod_counts = sample_totals - mod_counts
            # unmod_count = (sample_predictions != sample_mod_clusters).sum().item()
            cluster_counts[f'{sample_label}_mod'] = mod_counts.cpu().numpy().astype(np.uint32)
            cluster_counts[f'{sample_label}_unmod'] = unmod_counts.cpu().numpy().astype(np.uint32)

        return cluster_counts


    def retry(self, fn, delay=5, backoff=5, max_attempts=3, exception=Exception):
        attempts = 0
        while attempts < max_attempts - 1:
            try:
                return fn()
            except exception as e:
                attempts += 1
                self._worker.log("warning", f"Got error {e} on attempt {attempts}/{max_attempts}")
                time.sleep(delay)
                delay += backoff
        else:
            return fn()


def nanstd(X, dim):
    nonnans = (~X.isnan()).sum(dim)
    return (((X - X.nanmean(dim).unsqueeze(1)) ** 2).nansum(dim) / nonnans) ** 0.5


def cross_corr_matrix(pvalues, context=2):
    """
    Calculate the cross correlation matrix of the
    pvalues for a given context.
    """
    if len(pvalues) < (3*context) + 3:
        raise RuntimeError("Not enough p-values for a context of order %s"%context)

    pvalues = np.nan_to_num(pvalues, nan=1)
    if any(pvalues == 0) or any(np.isinf(pvalues)) or any(pvalues > 1):
        raise RuntimeError("At least one p-value is invalid")

    if all(pvalues == 1):
        return(np.ones((2*context + 1, 2*context + 1)))

    matrix = np.zeros((2*context + 1, 2*context + 1))
    s = pvalues.size
    for i in range(-context, context + 1):
        for j in range(-context, i + 1):
            x = i + context
            y = j + context
            matrix[x, y] = np.corrcoef(
                    np.roll(pvalues, i)[context:s-context],
                    np.roll(pvalues, j)[context:s-context])[0, 1]
    return matrix + np.tril(matrix, -1).T


def harmomic_series(sequence_context):
    return [1/(abs(i) + 1)
            for i in range(-sequence_context, sequence_context + 1)]


def combine_pvalues_hou(pvalues, weights, cor_mat):
    """ Hou's method for the approximation for the distribution of the weighted
        combination of non-independent or independent probabilities.
        If any pvalue is nan, returns nan.
        https://doi.org/10.1016/j.spl.2004.11.028
        pvalues: list of pvalues to be combined
        weights: the weights of the pvalues
        cor_mat: a matrix containing the correlation coefficients between pvalues
        Test: when weights are equal and cor=0, hou is the same as Fisher
        print(combine_pvalues([0.1,0.02,0.1,0.02,0.3], method='fisher')[1])
        print(hou([0.1,0.02,0.1,0.02,0.3], [1,1,1,1,1], np.zeros((5,5))))
    """
    if len(pvalues) != len(weights):
        raise ValueError("Can't combine pvalues if pvalues and weights are not the same length.")
    if cor_mat.shape[0] != cor_mat.shape[1] or cor_mat.shape[0] != len(pvalues):
        raise ValueError("The correlation matrix needs to be square, with each" + \
                         "dimension equal to the length of the pvalued vector.")
    if all(p == 1 for p in pvalues):
        return 1
    if any(p == 0 or np.isinf(p) or p > 1 for p in pvalues):
        raise ValueError("At least one p-value is invalid")

    # Covariance estimation as in Kost and McDermott (eq:8)
    # https://doi.org/10.1016/S0167-7152(02)00310-3
    cov = lambda r: (3.263*r) + (0.710*r**2) + (0.027*r**3)
    k = len(pvalues)
    cov_sum = np.float64(0)
    sw_sum = np.float64(0)
    w_sum = np.float64(0)
    tau = np.float64(0)
    for i in range(k):
        for j in range(i + 1, k):
            cov_sum += weights[i]*weights[j]*cov(cor_mat[i][j])
        sw_sum += weights[i]**2
        w_sum += weights[i]
        # Calculate the weighted Fisher's combination statistic
        tau += weights[i] * (-2*np.log(pvalues[i]))
    # Correction factor
    c = (2*sw_sum+cov_sum) / (2*w_sum)
    # Degrees of freedom
    f = (4*w_sum**2) / (2*sw_sum+cov_sum)
    # chi2.sf is the same as 1-chi2.cdf but is more accurate
    combined_p_value = chi2.sf(tau/c, f)
    # Return a very small number if pvalue = 0
    if combined_p_value == 0:
        combined_p_value = np.finfo(np.float64).tiny
    return combined_p_value


def calculate_lors(contingencies):
    odds1 = (contingencies[:, 0, 0]/contingencies[:, 0, 1])
    odds2 = (contingencies[:, 1, 0]/contingencies[:, 1, 1])
    return torch.round(torch.log(odds1/odds2), decimals=3)


def get_contigency_matrices(conditions, predictions):
    num_positions = predictions.shape[0]
    contingencies = torch.zeros((num_positions, 2, 2))
    expanded_conditions = conditions.unsqueeze(0).expand_as(predictions)
    combs = [(0, 0), (0, 1), (1, 0), (1, 1)]
    for comb in combs:
        cond_val, pred_val = comb
        comb_count = torch.logical_and(
                expanded_conditions == cond_val,
                predictions == pred_val
            ).sum(1)
        contingencies[:, cond_val, pred_val] = comb_count
    return contingencies




import re

from schema import Schema, And, Or, Optional

from nanocompore.common import is_valid_fasta
from nanocompore.common import Kit
from nanocompore.common import EVENTALIGN
from nanocompore.common import REMORA
from nanocompore.common import UNCALLED4


def valid_device(value):
    """
    Check if a device is valid.

    Parameters
    ----------
    value : str
        Device name

    Returns
    -------
        bool
    """
    return value in ['cpu', 'cuda'] or re.compile(r"cuda:\d+").match(value)


def validate_device(value):
    """
    Validate the device configuration.

    Parameters
    ----------
    value : dict
        Device configuration.

    Returns
    -------
        bool
    """
    if isinstance(value, dict):
        for k, v in value.items():
            if not valid_device(k) or not isinstance(v, int) or v < 0:
                return False
        return True
    raise ValueError("The value for 'devices' in the configuration "
                     f"is of unexpected type: {type(value)}. Should "
                     "be a map of devices and numbers of workers. "
                     'E.g. {"cpu": 5, "cuda:1": 8, "cuda:2": 8}')


def validate_db_data(config):
    """
    Validate if the input data contains input
    databases for all samples.

    Parameters
    ----------
    config : dict
        The configuration object.

    Returns
    -------
        bool
    """
    if config['resquiggler'] not in (EVENTALIGN, REMORA):
        return True
    for condition, cond_data in config['data'].items():
        for sample, sample_data in cond_data.items():
            if 'db' not in sample_data:
                return False
    return True


def validate_uncalled4_data(config):
    """
    Validate if the input data for Uncalled4 contains
    all necessary fields.

    Parameters
    ----------
    config : dict
        The configuration object.

    Returns
    -------
        bool
    """
    if config['resquiggler'] != UNCALLED4:
        return True
    for condition, cond_data in config['data'].items():
        for sample, sample_data in cond_data.items():
            if 'bam' not in sample_data:
                return False
    return True


def validate_auto_test_multiple_test_correcton(config):
    """
    Validate that either Benjamini-Hochberg or Bonferroni
    is used for correcting for multiple tests when we use
    the auto test. We haven't implemented support for other
    methods.

    Parameters
    ----------
    config : dict
        The configuration object.

    Returns
    -------
        bool
    """
    tests = config.get('comparison_methods', DEFAULT_COMPARISON_METHODS)
    if 'auto' in tests or 'AUTO' in tests:
        correction_method = config.get('correction_method',
                                       DEFAULT_CORRECTION_METHOD)
        return correction_method in ('fdr_bh', 'bonferroni')
    return True


def depleted_condition_exists(config):
    return config['depleted_condition'] in config['data']


CONFIG_SCHEMA = Schema(And({
    'data': And(
        {
            str: { # condition
                str: { # sample (replicate)
                    Optional('bam'): lambda f: open(f, 'r'),
                    Optional('db'): str,
                }
            }
        },
        And(lambda d: len(d) == 2, error='Only two conditions allowed')
    ),
    'depleted_condition': str,
    'fasta': And(is_valid_fasta, error='Invalid fasta file'),
    'resquiggler': Or('uncalled4', 'eventalign', 'remora'),
    'kit': Or(*[v.name for v in Kit]),
    Optional('devices'): validate_device,
    Optional('gtf'): And(lambda f: open(f, 'r'), error='Invalid GTF file'),
    Optional('min_coverage'): And(int, lambda n: n >= 0, error='min_coverage must be >= 0'),
    Optional('downsample_high_coverage'): And(int, lambda n: n >= 0, error='downsample_high_coverage must be >= 0'),
    Optional('min_ref_length'): And(int, lambda n: n >= 0, error='min_ref_length must be >= 0'),
    Optional('max_invalid_kmers_freq'): And(float, lambda n: n >= 0, error='max_invalid_kmers_freq must be in the [0, 1] range'),
    Optional('selected_refs'): And(lambda f: open(f, 'r'), error='Invalid references file.'),
    Optional('ignored_refs'): And(lambda f: open(f, 'r'), error='Invalid references file.'),
    Optional('comparison_methods'): And(['GMM',
                                         'GOF',
                                         'KS',
                                         'TT',
                                         'MW',
                                         'auto',
                                         'AUTO',
                                         'GAUSSIAN_MIXTURE_MODEL',
                                         'GOODNESS_OF_FIT',
                                         'KOLMOGOROV_SMIRNOV',
                                         'T_TEST',
                                         'MANN_WHITNEY'],
                                        lambda l: len(l) > 0),
    Optional('motor_dwell_offset'): And(int, lambda n: n >= 0, error='motor_dwell_offset must be >= 0'),
    Optional('sequence_context'): And(int, lambda n: n >= 0 and n <= 4, error='sequence_context must be >= 0 and <= 4'),
    Optional('sequence_context_weights'): Or('uniform', 'harmonic'),
    Optional('outpath'): str,
    Optional('outprefix'): str,
    Optional('result_exists_strategy'): Or("stop", "continue", "overwrite"),
    Optional('log_level'): Or('warning', 'info', 'debug'),
    Optional('progress'): bool,
    Optional('export_shift_stats'): bool,
    Optional('correction_method'): 'fdr_bh'},
    # Additional validation of the full configuration
    And(validate_db_data,
        error='When using the "eventalign" or "remora" resquigglers ' +
              'each sample in the configuration must contain ' +
              'a "db" entry with the path to an SQLite DB ' +
              'produced by the eventalign_collapse subcommand.'),
    And(validate_uncalled4_data,
        error='When using the "uncalled4" resquiggler ' +
              'each sample in the configuration must contain ' +
              'the field "bam" with a path to the aligned bam file ' +
              'that contains the resquiggling tags produced ' +
              'by Uncalled4.'),
    And(depleted_condition_exists,
        error="The condition set in 'depleted_condition' is not " + \
              "defined in 'data'."),
    And(validate_auto_test_multiple_test_correcton,
        error="If the auto test is used in 'comparison_methods' then 'correction_method' "
              "should be either 'fhr_bh' (for Benjamini-Hochberg) or 'bonferroni'.")))


DEFAULT_KIT = 'RNA002'
DEFAULT_DEVICES = {'cpu': 2}
DEFAULT_MIN_COVERAGE = 30
DEFAULT_MAX_READS = 5000
DEFAULT_DOWNSAMPLE_HIGH_COVERAGE = 5000
DEFAULT_MIN_REF_LENGTH = 100
DEFAULT_MAX_INVALID_KMERS_FREQ = 0.1
DEFAULT_COMPARISON_METHODS = ['GMM', 'KS']
DEFAULT_MOTOR_DWELL_OFFSET = 0
DEFAULT_SEQUENCE_CONTEXT = 0
DEFAULT_SEQUENCE_CONTEXT_WEIGHTS = 'uniform'
DEFAULT_OUTPATH = 'nanocompore_output'
DEFAULT_OUTPREFIX = 'out_'
DEFAULT_RESULT_EXISTS_STRATEGY = 'stop'
DEFAULT_LOG_LEVEL = 'info'
DEFAULT_PROGRESS = False
DEFAULT_EXPORT_SHIFT_STATS = False
DEFAULT_CORRECTION_METHOD = 'fdr_bh'


class Config:
    """
    Represents the input configuration for a Nanocompore experiment.
    An instance contains information about the input data files and
    any additoinal parameters for Nanocompore.
    """

    def __init__(self, config_file):
        self._config = CONFIG_SCHEMA.validate(config_file)

        self._test_condition = [cond
                                for cond in self.get_condition_labels()
                                if cond != self.get_depleted_condition()][0]


    def get_data(self):
        return self._config['data']


    def get_resquiggler(self):
        """
        Returns the resquiggler software specified
        by the user.
        """
        return self._config['resquiggler']


    def get_kit(self):
        """
        Returns an instance of the Kit enum representing
        the sequencing kit, e.g. RNA002.
        """
        kit_name = self._config.get('kit', DEFAULT_KIT)
        return Kit[kit_name]


    def get_devices(self):
        """
        Returns the device to be used for computations.
        E.g. cpu or cuda.
        """
        return self._config.get('devices', DEFAULT_DEVICES)


    def get_fasta_ref(self):
        """
        Reference fasta file that was used for mapping the data.
        """
        return self._config['fasta']


    def get_gtf(self):
        """
        GTF file with annotation of transcriptome used for mapping.
        """
        return self._config.get('gtf')


    def get_min_coverage(self):
        """
        Minimum coverage required in each condition to do the comparison.
        """
        return self._config.get('min_coverage', DEFAULT_MIN_COVERAGE)


    def get_downsample_high_coverage(self):
        """
        Transcripts with high coverage will be downsampled.
        """
        return self._config.get('downsample_high_coverage', DEFAULT_DOWNSAMPLE_HIGH_COVERAGE)


    def get_min_ref_length(self):
        """
        Minimum length of a reference transcript to include it in the analysis.
        """
        return self._config.get('min_ref_length', DEFAULT_MIN_REF_LENGTH)


    def get_max_invalid_kmers_freq(self):
        """
        Maximum allowed ratio of invalid kmers in the read.
        """
        return self._config.get('max_invalid_kmers_freq', DEFAULT_MAX_INVALID_KMERS_FREQ)


    def get_selected_refs(self):
        """
        Text file with one reference per line.
        Only references found in the file will be analysed.
        """
        return self._config.get('selected_refs')


    def get_ignored_refs(self):
        """
        Text file with one reference per line.
        References found within the file will not be analysed.
        """
        return self._config.get('ignored_refs')


    def get_comparison_methods(self):
        """
        List of comparison methods. Valid methods are: GMM,KS,TT,MW.
        """
        remappings = {
            'GAUSSIAN_MIXTURE_MODEL': 'GMM',
            'KOLMOGOROV_SMIRNOV': 'KS',
            'T_TEST': 'TT',
            'MANN_WHITNEY': 'MW',
        }
        return [remappings.get(m, m)
                for m in self._config.get('comparison_methods', DEFAULT_COMPARISON_METHODS)]


    def get_motor_dwell_offset(self):
        """
        Interactions between a modification and the motor protein
        can affect the translocation speed. If present, this
        should affect the dwell time of the downstream position.
        This returns the expected offset for this effect.
        """
        return self._config.get('motor_dwell_offset', DEFAULT_MOTOR_DWELL_OFFSET)


    def get_sequence_context(self):
        """
        Sequence context for combining p-values.
        """
        return self._config.get('sequence_context', DEFAULT_SEQUENCE_CONTEXT)


    def get_sequence_context_weights(self):
        """
        Type of weights to use for combining p-values.
        """
        return self._config.get('sequence_context_weights', DEFAULT_SEQUENCE_CONTEXT_WEIGHTS)


    def get_outpath(self):
        """
        Path to the output folder.
        """
        return self._config.get('outpath', DEFAULT_OUTPATH)


    def get_outprefix(self):
        """
        Text outprefix for all the files generated.
        """
        prefix = self._config.get('outprefix', DEFAULT_OUTPREFIX)
        if prefix and prefix.endswith('_'):
            return prefix
        return prefix + '_'


    def get_result_exists_strategy(self):
        """
        What to do if <outpath> already exists.
        Options are: stop, continue, and overwrite.
        Default: stop
        """
        return self._config.get('result_exists_strategy', DEFAULT_RESULT_EXISTS_STRATEGY)


    def get_log_level(self):
        """
        Get the log level.
        """
        return self._config.get('log_level', DEFAULT_LOG_LEVEL).upper()


    def get_progress(self):
        """
        Display a progress bar during execution.
        """
        return self._config.get('progress', DEFAULT_PROGRESS)


    def get_export_shift_stats(self):
        """
        Whether to export shift statistics from the DB
        to a TSV during postprocessing.
        """
        return self._config.get('export_shift_stats', DEFAULT_EXPORT_SHIFT_STATS)


    def get_correction_method(self):
        """
        Get the multiple test correction method.
        """
        return self._config.get('correction_method', DEFAULT_CORRECTION_METHOD)


    def get_sample_labels(self):
        return [sample
                for samples in self.get_data().values()
                for sample in samples]


    def get_condition_labels(self):
        return list(self.get_data().keys())


    def get_depleted_condition(self):
        return self._config['depleted_condition']


    def get_test_condition(self):
        return self._test_condition


    def is_multi_replicate(self):
        return all(len(reps) > 1
                   for reps in self.get_data().values())


    def get_sample_ids(self):
        labels = self.get_sample_labels()
        return dict(zip(labels, range(len(labels))))


    def get_condition_ids(self) -> dict[str, int]:
        """
        Returns a (condition label => condition id)
        mapping.

        Returns
        -------
        dict[str, int]
            Condition label => condition id dict
        """
        labels = self.get_condition_labels()
        return dict(zip(labels, range(len(labels))))


    def sample_to_condition(self):
        return {sample: cond
                for cond, sample_defs in self.get_data().items()
                for sample in sample_defs.keys()}


    def get_condition_samples(self, condition_label):
        return list(self.get_data()[condition_label].keys())


    def get_conditions_to_samples(self):
        return {cond: list(sample_defs.keys())
                for cond, sample_defs in self.get_data().items()}


    def get_sample_condition_bam_data(self):
        return [(sample, condition, samp_def['bam'])
                for condition, samples in self.get_data().items()
                for sample, samp_def in samples.items()]


    def get_sample_condition_db_data(self):
        return [(sample, condition, samp_def['db'])
                for condition, samples in self.get_data().items()
                for sample, samp_def in samples.items()]


    def get_sample_pod5_bam_data(self):
        return [(sample, samp_def['pod5'], samp_def['bam'])
                for _, samples in self.get_data().items()
                for sample, samp_def in samples.items()]




import pytest
import schema

import copy
import os

from nanocompore.config import Config
from tests.common import BASIC_CONFIG, cwd


class TestConfig:
    def test_valid_config(self):
       yaml = BASIC_CONFIG

       config = Config(yaml)

       assert config.get_fasta_ref() == os.path.join(cwd, 'tests/fixtures/test_reference.fa')
       assert config.get_data() == yaml['data']


    def test_too_many_conditions(self):
        yaml = copy.deepcopy(BASIC_CONFIG)
        yaml['data']['third_condition'] = yaml['data']['kd']

        with pytest.raises(schema.SchemaError) as exception_info:
            config = Config(yaml)

        assert 'Only two conditions allowed' == str(exception_info.value)


    def test_too_few_conditions(self):
        yaml = copy.deepcopy(BASIC_CONFIG)
        del yaml['data']['kd']

        with pytest.raises(schema.SchemaError) as exception_info:
            config = Config(yaml)

        assert 'Only two conditions allowed' == str(exception_info.value)


    def test_missing_fasta(self):
        yaml = copy.deepcopy(BASIC_CONFIG)
        del yaml['fasta']

        with pytest.raises(schema.SchemaError) as exception_info:
            config = Config(yaml)

        assert "Missing key: 'fasta'" == str(exception_info.value)


    def test_invalid_fasta(self):
        yaml = copy.deepcopy(BASIC_CONFIG)
        yaml['fasta'] = yaml['data']['kd']['kd1']['bam']

        with pytest.raises(schema.SchemaError) as exception_info:
            config = Config(yaml)

        assert "Invalid fasta file" == str(exception_info.value)


    def test_unsupported_comparison_method(self):
        yaml = copy.deepcopy(BASIC_CONFIG)
        yaml['comparison_methods'] = ['unsupported_method']

        with pytest.raises(schema.SchemaError) as exception_info:
            config = Config(yaml)

        assert "did not validate 'unsupported_method'" in str(exception_info.value)


    def test_comparison_method_is_case_sensitive(self):
        yaml = copy.deepcopy(BASIC_CONFIG)
        yaml['comparison_methods'] = ['gmm']

        with pytest.raises(schema.SchemaError) as exception_info:
            config = Config(yaml)

        assert "did not validate 'gmm'" in str(exception_info.value)


    def test_incorrect_sequence_context(self):
        yaml = copy.deepcopy(BASIC_CONFIG)
        yaml['sequence_context'] = 5

        with pytest.raises(schema.SchemaError) as exception_info:
            config = Config(yaml)

        assert "sequence_context must be >= 0 and <= 4" == str(exception_info.value)


    def test_unsupported_correction_method(self):
        yaml = copy.deepcopy(BASIC_CONFIG)
        yaml['correction_method'] = 'bonferroni'

        with pytest.raises(schema.SchemaError) as exception_info:
            config = Config(yaml)

        assert "does not match 'bonferroni'" in str(exception_info.value)


    def test_missing_kit(self):
        yaml = copy.deepcopy(BASIC_CONFIG)
        del yaml['kit']

        with pytest.raises(schema.SchemaError) as exception_info:
            config = Config(yaml)

        assert "Missing key: 'kit'" in str(exception_info.value)


    def test_unsupported_kit(self):
        yaml = copy.deepcopy(BASIC_CONFIG)
        yaml['kit'] = 'RNA500'

        with pytest.raises(schema.SchemaError) as exception_info:
            config = Config(yaml)

        assert "did not validate 'RNA500'" in str(exception_info.value)


    def test_missing_resquiggler(self):
        yaml = copy.deepcopy(BASIC_CONFIG)
        del yaml['resquiggler']

        with pytest.raises(schema.SchemaError) as exception_info:
            config = Config(yaml)

        assert "Missing key: 'resquiggler'" in str(exception_info.value)


    def test_missing_depleted_condition(self):
        yaml = copy.deepcopy(BASIC_CONFIG)
        del yaml['depleted_condition']

        with pytest.raises(schema.SchemaError) as exception_info:
            config = Config(yaml)

        assert "Missing key: 'depleted_condition'" in str(exception_info.value)


    def test_nonexisting_depleted_condition(self):
        yaml = copy.deepcopy(BASIC_CONFIG)
        yaml['depleted_condition'] = 'does_not_exist'

        with pytest.raises(schema.SchemaError) as exception_info:
            config = Config(yaml)


    def test_bam_required_for_uncalled4(self):
        yaml = copy.deepcopy(BASIC_CONFIG)
        del yaml['data']['kd']['kd1']['bam']

        with pytest.raises(schema.SchemaError) as exception_info:
            config = Config(yaml)


    def test_db_required_for_eventalign(self):
        yaml = copy.deepcopy(BASIC_CONFIG)
        yaml['resquiggler'] = 'eventalign'

        # error when missing both
        with pytest.raises(schema.SchemaError) as exception_info:
            config = Config(yaml)

        # should be okay if eventalign_db is set
        for cond in yaml['data']:
            for sample in yaml['data'][cond]:
                yaml['data'][cond][sample]['db'] = 'some/path'
        config = Config(yaml)


    def test_db_required_for_remora(self):
        yaml = copy.deepcopy(BASIC_CONFIG)
        yaml['resquiggler'] = 'remora'

        # error when missing both
        with pytest.raises(schema.SchemaError) as exception_info:
            config = Config(yaml)

        # should be okay if eventalign_db is set
        for cond in yaml['data']:
            for sample in yaml['data'][cond]:
                yaml['data'][cond][sample]['db'] = 'some/path'
        config = Config(yaml)




from typing import Optional

import numpy as np
import pod5
import pysam

from jaxtyping import Float
from loguru import logger
from pkg_resources import resource_filename
from remora import io, refine_signal_map, RemoraError

from nanocompore.common import Kit
from nanocompore.common import NanocomporeError


RNA002_LEVELS_FILE = "models/rna002_5mer_levels_v1.txt"
RNA004_LEVELS_FILE = "models/rna004_9mer_levels_v1.txt"
VAR_ORDER = ['dwell', 'trimmean', 'trimsd']

DEFAULT_MAX_READS = 5000


class Remora:
    """
    Resquiggles reads using ONT's Remora.
    """

    def __init__(self,
                 pod5_path: str,
                 bam_path: str,
                 kit: Optional[Kit]=Kit.RNA004,
                 max_reads: Optional[int]=DEFAULT_MAX_READS):
        """

        Parameters
        ----------
        kit : Optional[Kit]
            The sequencing kit used. By default will use RNA004.

        max_reads : Optional[int]
            Maximum number of reads to resquiggle.
            By default will use no more than 5000.

        pod5_path : str
            The path to the pod5 file with the signal data.

        bam_path : str
            The path to the aligned bam file.
        """
        self._kit = kit
        self._max_reads = max_reads
        try:
            self._pod5 = pod5.Reader(pod5_path)
        except Exception:
            raise NanocomporeError(f"Failed to open pod5 file {pod5}")

        try:
            self._bam = pysam.AlignmentFile(bam_path)
        except Exception:
            raise NanocomporeError(f"Failed to open bam file {bam_path}")

        # Remora requires a kmer model file to resquiggle
        # the data (signal to sequence alignment). This is
        # defined as a singal refiner object in the Remora API.
        # Without the signal refiner, it will not resquiggle,
        # but instead merely return the ionic current stream
        try:
            self._sig_map_refiner = self._check_signal_refiner(kit=kit)
        except Exception:
            raise NanocomporeError("Failed to create the signal map refiner. "
                                   "Check that the kmer model table is up-to-date")

        # TODO test Nanocompore accuracy using seconds per kmer
        # or samples per kmer. Remora returns the number of
        # datapoints per kmer, not the number of seconds the
        # kmer persisted in the sensitive region of the nanopore.
        # This function uses the pod5 api to convert the sampling
        # rate of the sequencing (hz) to time per sample (seconds).
        try:
            self._time_per_sample = self._get_time_per_sample()
        except Exception:
            raise NanocomporeError("Failed to check for sampling rate. "
                                   "Likely something wrong with the pod5 file")


    def get_resquiggled_data(
            self,
            ref_id:str,
            ref_seq:str
    ) -> tuple[Float[np.ndarray, "reads positions"],
               Float[np.ndarray, "reads positions"],
               list[str]]:
        """
        Returns the intensity and dwell time data
        for the resquiggled reads.

        Parameters
        ----------
        ref_id : str
            Reference ID of the transcript.
        ref_seq :
            Reference sequence of the transcript.

        Returns
        -------
        tuple[Float[np.ndarray, "reads positions"],
              Float[np.ndarray, "reads positions"],
              list[str]]
            Tuple with:
              - 2D array with shape (reads, positions)
                containing the intensity values.
              - 2D array with shape (reads, positions)
                containing the dwell time values.
              - list with the qname ids of the reads
        """
        ref_region = io.RefRegion(ctg=ref_id,
                                  strand='+',
                                  start=0,
                                  end=len(ref_seq))

        try:
            # Resquiggle the signal and get the summary metrics for
            # each position of the transcript. The result is a dict
            # with metrics:
            # {metric: <numpy appray with shape (reads, positions)>, ...}
            samples_metrics, bam_reads = self._remora_resquiggle(ref_region)
            return (samples_metrics['trimmean'],
                    samples_metrics['dwell'],
                    bam_reads)
        except RemoraError as e:
            if str(e) == "No reads covering region":
                return None
            raise NanocomporeError("failed to resquiggle with Remora") from e
        except Exception as e:
            raise NanocomporeError("failed to resquiggle with Remora") from e


    @property
    def kmer_size(self):
        return self._kmer_size


    def _check_signal_refiner(self, kit):
        level_table = self._kmer_model_selector(kit=kit)
        self._kmer_size = self._kmer_size_detector(level_table=level_table)

        sig_map_refiner = refine_signal_map.SigMapRefiner(
            kmer_model_filename=level_table,
            scale_iters=0,
            algo='dwell_penalty',
            do_rough_rescale=True,
            do_fix_guage=True,
        )
        logger.debug("sig_map_refiner properly opened")
        return sig_map_refiner


    def _get_time_per_sample(self):
        logger.trace("Attempting to calculate time per sample")
        read = next(self._pod5.reads())
        sample_rate = read.run_info.sample_rate
        logger.trace(f"Sampling rate is {sample_rate} Hz")
        time_per_sample = 1.0/sample_rate
        logger.trace(f"Time per sample is {time_per_sample} seconds")
        return time_per_sample


    def _remora_resquiggle(self, ref_region):
        logger.debug(f"Starting to resquiggle data with Remora API for {ref_region.ctg}")
        samples_metrics, all_bam_reads = io.get_ref_reg_samples_metrics(
            ref_region,
            [(self._pod5, self._bam)],
            metric="dwell_trimmean_trimsd",
            sig_map_refiner=self._sig_map_refiner,
            max_reads=self._max_reads,
            # we only support directRNA for now which is sequenced in 3'->5' direction
            reverse_signal=True,
            signal_type='norm',
        )
        logger.debug(f"Data for {ref_region.ctg} resquiggled")
        return samples_metrics[0], [read.qname for read in all_bam_reads[0]]


    def _kmer_model_selector(self, kit):
        if kit == Kit.RNA002:
            level_table = resource_filename('nanocompore', RNA002_LEVELS_FILE)
        elif kit == Kit.RNA004:
            level_table = resource_filename('nanocompore', RNA004_LEVELS_FILE)
        else:
            raise NotImplementedError(f"Kit {kit} not implemented yet.")

        return level_table


    def _kmer_size_detector(self, level_table):
        with open(level_table, 'r') as infile:
            line = infile.readline().strip().split('\t')
            kmer_size = len(line[0])
        logger.debug(f'The kmer size is {kmer_size}')
        return kmer_size




import copy
import sqlite3

from pathlib import Path

import numpy as np

from nanocompore.common import READ_ID_TYPE
from nanocompore.common import MEASUREMENTS_TYPE
from nanocompore.common import Kit
from nanocompore.config import Config
from nanocompore.preprocessing import RemoraPreprocessor
from tests.common import BASIC_CONFIG


def test_remora_preprocessing():
    """
    Integration test for RemoraPreprocessor
    """
    output = 'kd1_remora.sqlite'
    preprocessor = RemoraPreprocessor(BASIC_CONFIG['fasta'],
                                      'tests/fixtures/kd1.pod5',
                                      'tests/fixtures/kd1.bam',
                                      output,
                                      Kit['RNA002'],
                                      5000,
                                      2)

    try:
        preprocessor()

        conn = sqlite3.connect(output)

        num_transcripts = conn.execute('SELECT COUNT(*) FROM transcripts').fetchone()[0]
        assert num_transcripts == 2

        signal_data = conn.execute('SELECT * FROM signal_data').fetchall()
        assert len(signal_data) > 0

        # Validate that we have data for two transcripts
        assert {1, 2} == {row[0] for row in signal_data}

        # Validate that all reads are included in the
        # preprocessing database.
        read_ids = {row[1] for row in signal_data}
        query = f'''SELECT read
                    FROM reads
                    WHERE id IN ({",".join(["?" for _ in range(len(read_ids))])})'''
        read_uuids = conn.execute(query, tuple(read_ids)).fetchall()
        read_uuids = {row[0] for row in read_uuids}

        assert read_uuids == {
                '3f46f499-8ce4-4817-8177-8ad61b784f27',
                '73d62df4-f04a-4207-a4bc-7b9739b3c3b2',
                'b7bc9a36-318e-4be2-a90f-74a5aa6439bf',
                '65db1ced-cbf5-40e4-880b-9f202c715804',
                '6384da30-e6cc-4421-b93a-3aca69d09857',
                '6c80b542-3176-49c8-b6df-a204499cc0dd'}


        get_actb_id = 'SELECT id FROM transcripts WHERE name LIKE "%ENST00000674681.1%"'
        transcript_id = conn.execute(get_actb_id).fetchone()[0]

        # get the intensity for all reads for ACTB
        reads = np.array([np.frombuffer(row[2], dtype=MEASUREMENTS_TYPE)
                          for row in signal_data
                          if row[0] == transcript_id])

        pos301 = reads[:, 301]
        assert round(float(pos301.mean()), 3) == 0.806
    finally:
        Path(output).unlink()




import os
import sqlite3
import json

from contextlib import closing
from pathlib import Path
from typing import Iterator, Union

import pandas as pd
import numpy as np

from loguru import logger

from nanocompore.common import NanocomporeError
from nanocompore.common import TranscriptRow


# ======= RESULT DATABASE QUERIES =======

CREATE_TRANSCRIPTS_RESULTS_TABLE = """
CREATE TABLE IF NOT EXISTS transcripts (
    id INTEGER PRIMARY KEY AUTOINCREMENT,
    name VARCHAR NOT NULL UNIQUE
);
"""
INSERT_TRANSCRIPTS_QUERY = """
INSERT INTO transcripts (id, name) VALUES(?, ?);
"""

CREATE_KMER_RESULTS_TABLE = """
CREATE TABLE IF NOT EXISTS kmer_results (
    id INTEGER PRIMARY KEY AUTOINCREMENT,
    transcript_id INTEGER NOT NULL,
    pos INTEGER NOT NULL,
    kmer INTEGER NOT NULL,
    UNIQUE (transcript_id, pos),
    FOREIGN KEY (transcript_id) REFERENCES transcripts(id)
);
"""

CREATE_KMER_RESULTS_TRANSCRIPT_ID_INDEX = """
CREATE INDEX IF NOT EXISTS kmer_results_transcript_id_index
    ON kmer_results(transcript_id);
"""

CREATE_TRANSCRIPTS_NAME_INDEX = """
CREATE INDEX IF NOT EXISTS transcripts_name_index
    ON transcripts(name);
"""

# ======= PREPROCESSING DATABASES QUERIES =======

DROP_METADATA_TABLE_QUERY = """
DROP TABLE IF EXISTS metadata;
"""

CREATE_METADATA_TABLE_QUERY = """
CREATE TABLE metadata (
    key TEXT NOT NULL,
    value TEXT NOT NULL,
    PRIMARY KEY(key)
);
"""

DROP_SIGNAL_DATA_TABLE_QUERY = """
DROP TABLE IF EXISTS signal_data;
"""

CREATE_SIGNAL_DATA_TABLE_QUERY = """
CREATE TABLE signal_data (
    transcript_id INTEGER NOT NULL,
    read_id INTEGER NOT NULL,
    intensity BLOB,
    dwell BLOB
);
"""

INSERT_SIGNAL_DATA_QUERY = """
INSERT INTO signal_data VALUES (?, ?, ?, ?);
"""

DROP_READS_TABLE_QUERY = """
DROP TABLE IF EXISTS reads;
"""

CREATE_READS_TABLE_QUERY = """
CREATE TABLE reads (
    read TEXT NOT NULL,
    id INTEGER NOT NULL,
    invalid_kmers REAL
);
"""

DROP_READS_ID_INDEX_QUERY = """
DROP INDEX IF EXISTS reads_id_index;
"""

CREATE_READS_ID_INDEX_QUERY = """
CREATE INDEX IF NOT EXISTS reads_id_index ON reads (id);
"""

DROP_SIGNAL_DATA_INDEX_QUERY = """
DROP INDEX IF EXISTS signal_data_index;
"""

CREATE_SIGNAL_DATA_INDEX_QUERY = """
CREATE INDEX IF NOT EXISTS signal_data_index ON signal_data (transcript_id, read_id);
"""

DROP_TRANSCRIPTS_TABLE_QUERY = """
DROP TABLE IF EXISTS transcripts;
"""

CREATE_TRANSCRIPTS_TABLE_QUERY = """
CREATE TABLE transcripts (
    name TEXT NOT NULL,
    id INTEGER NOT NULL,
    PRIMARY KEY(name)
);
"""

DROP_TRANSCRIPTS_ID_INDEX_QUERY = """
DROP INDEX IF EXISTS transcripts_id_index;
"""

CREATE_TRANSCRIPTS_ID_INDEX_QUERY = """
CREATE INDEX IF NOT EXISTS transcripts_id_index ON transcripts (id);
"""

INSERT_READS_QUERY = """
INSERT INTO reads (read, id, invalid_kmers) VALUES(?, ?, ?);
"""

GET_SIGNAL_DATA_FOR_TRANSCRIPT_QUERY = """
SELECT intensity, dwell
FROM signal_data sd
JOIN reads r ON sd.read_id = r.id
JOIN transcripts t ON sd.transcript_id = t.id
WHERE t.name = ?
  AND r.invalid_kmers <= ?
"""

BASE_KMER_RESULT_COLUMNS = ['id', 'transcript_id', 'pos', 'kmer']


class ResultsDB():
    def __init__ (self, config, init_db=False):
        self._config = config

        if config.get_outpath():
            self._outpath = config.get_outpath()
        else:
            self._outpath = os.getcwd()

        self._prefix = config.get_outprefix()
        self.db_path = os.path.join(self._outpath, f"{self._prefix}sampComp_sql.db")
        if init_db:
            self._setup_database(config.get_result_exists_strategy())
            self._create_tables()


    def get_results(self,
                    columns: Union[list[str], None]=None,
                    chunksize: Union[int, None]=None) -> Union[pd.DataFrame, Iterator[pd.DataFrame]]:
        if not columns:
            cols = '*'
        else:
            cols = ','.join(columns)
        with closing(sqlite3.connect(self.db_path)) as conn:
            query = f"""
            SELECT {cols}
            FROM kmer_results res
            JOIN transcripts t ON t.id = res.transcript_id
            """
            for chunk in pd.read_sql(query, conn, chunksize=chunksize):
                yield chunk


    def get_all_results(self):
        with closing(sqlite3.connect(self.db_path)) as conn:
            query = """
            SELECT *
            FROM kmer_results res
            JOIN transcripts t ON t.id = res.transcript_id
            """
            return pd.read_sql(query, conn)


    def get_transcripts(self):
        with closing(sqlite3.connect(self.db_path)) as conn,\
             closing(conn.cursor()) as cursor:
            query = "SELECT name FROM transcripts"
            return [row[0] for row in cursor.execute(query).fetchall()]


    def get_result_column_names(self) -> list[str]:
        with closing(sqlite3.connect(self.db_path)) as conn,\
             closing(conn.cursor()) as cursor:
            res = cursor.execute('SELECT * FROM kmer_results LIMIT 1')
            return [description[0] for description in res.description]


    def get_columns(self, cols: list[str]) -> pd.DataFrame:
        col_list = ', '.join(cols)
        with closing(sqlite3.connect(self.db_path)) as conn:
            query = f"""
            SELECT {col_list}
            FROM kmer_results res
            JOIN transcripts t ON t.id = res.transcript_id
            """
            return pd.read_sql(query, conn)


    def get_columns_for_ref(self, cols: list[str], ref: str) -> pd.DataFrame:
        """
        Returns the data for the given reference, selecting
        only the listed columns.

        Parameters
        ----------
        cols : list[str]
            List of column names to select.
        ref : str
            Reference id.

        Returns
        -------
        pd.DataFrame
            The dataframe would contain a "pos" column
            along with the list of provided columns.
        """
        cols = ['pos'] + cols
        col_list = ', '.join(cols)
        query = f'''
        SELECT {col_list}
        FROM kmer_results res
        JOIN transcripts t ON t.id = res.transcript_id
        WHERE t.name = ?
        '''
        with closing(sqlite3.connect(self.db_path)) as conn,\
             closing(conn.cursor()) as cursor:
            results = cursor.execute(query, (ref,)).fetchall()

            df = pd.DataFrame([dict(zip(cols, row))
                               for row in results])
            return df


    def get_next_transcript_id(self):
        with closing(sqlite3.connect(self.db_path)) as conn,\
             closing(conn.cursor()) as cursor:
            query = "SELECT MAX(id) FROM transcripts"
            return cursor.execute(query).fetchone()[0] + 1


    def save_test_results(self, transcript, test_results):
        with closing(sqlite3.connect(self.db_path)) as conn,\
             closing(conn.cursor()) as cursor:
            cursor.execute(INSERT_TRANSCRIPTS_QUERY, (transcript.id, transcript.name))
            test_columns = dict(zip(test_results.columns, test_results.dtypes))
            self._create_missing_columns(test_columns, cursor)
            test_results.to_sql('kmer_results', conn, if_exists='append', index=False)


    def save_transcript(self, transcript):
        with closing(sqlite3.connect(self.db_path)) as conn,\
             closing(conn.cursor()) as cursor:
            cursor.execute(INSERT_TRANSCRIPTS_QUERY, (transcript.id, transcript.name))
            conn.commit()


    def index_database(self):
        with closing(sqlite3.connect(self.db_path)) as conn,\
             closing(conn.cursor()) as cursor:
            cursor.execute(CREATE_TRANSCRIPTS_NAME_INDEX)
            cursor.execute(CREATE_KMER_RESULTS_TRANSCRIPT_ID_INDEX)


    def _create_missing_columns(self, test_columns, cursor):
        info = cursor.execute("SELECT * FROM kmer_results LIMIT 1")
        existing_columns = {item[0] for item in info.description}
        logger.trace("Adding columns to table kmer_results")
        try:
            for column, column_type in test_columns.items():
                if column in existing_columns:
                    continue

                if column_type in (float, np.float64, np.float32):
                    column_type = 'FLOAT'
                elif column_type is str:
                    column_type = 'VARCHR'
                elif column_type is int:
                    column_type = 'INTEGER'
                else:
                    column_type = 'VARCHAR'

                update_query = f"ALTER TABLE kmer_results ADD COLUMN {column} {column_type}"
                cursor.execute(update_query)
                existing_columns.add(column)
                logger.trace(f"Added {column} to kmer_results")
        except Exception:
            raise NanocomporeError(f"Database error: Failed to insert at least one of {test_columns} new labels into kmer_results of {self.db_path}")


    def _setup_database(self, result_exists_strategy):
        if os.path.isfile(self.db_path):
            if result_exists_strategy == 'overwrite':
                os.remove(self.db_path)
                logger.debug(f"Removed existing database file '{self.db_path}'")
            elif result_exists_strategy == 'continue':
                logger.info(f"Database file '{self.db_path}' already exists and result_exists_strategy is set to 'continue'. Will try to reuse it.")
            else:
                raise NanocomporeError(f"Database file '{self.db_path}' exists and 'results_exists_strategy' is 'stop'")
        with closing(sqlite3.connect(self.db_path)) as conn:
            conn.execute('PRAGMA foreign_keys = ON')
            conn.execute('PRAGMA journal_mode = wal')
            conn.execute('PRAGMA synchronous = NORMAL')


    def _create_tables(self):
        with closing(sqlite3.connect(self.db_path)) as conn,\
             closing(conn.cursor()) as cursor:
            cursor.execute(CREATE_TRANSCRIPTS_RESULTS_TABLE)
            cursor.execute(CREATE_KMER_RESULTS_TABLE)


class PreprocessingDB:
    def __init__(self, db):
        self._db = db


    def connect(self):
        conn = sqlite3.connect(self._db, isolation_level=None)
        conn.execute('PRAGMA synchronous = OFF')
        conn.execute('PRAGMA journal_mode = OFF')
        # Use largest possible page size of 65kb
        conn.execute('PRAGMA page_size = 65536')
        # Use cache of 2048 x page_size. That's 128MB of cache.
        conn.execute('PRAGMA cache_size = 2000')
        return conn


    def setup(self, metadata):
        with closing(self.connect()) as conn,\
             closing(conn.cursor()) as cursor:
            cursor.execute(DROP_READS_ID_INDEX_QUERY)
            cursor.execute(DROP_SIGNAL_DATA_INDEX_QUERY)
            cursor.execute(DROP_SIGNAL_DATA_TABLE_QUERY)
            cursor.execute(CREATE_SIGNAL_DATA_TABLE_QUERY)
            cursor.execute(DROP_READS_TABLE_QUERY)
            cursor.execute(CREATE_READS_TABLE_QUERY)
            cursor.execute(DROP_TRANSCRIPTS_TABLE_QUERY)
            cursor.execute(CREATE_TRANSCRIPTS_TABLE_QUERY)
            cursor.execute(DROP_METADATA_TABLE_QUERY)
            cursor.execute(CREATE_METADATA_TABLE_QUERY)
            cursor.execute(DROP_TRANSCRIPTS_ID_INDEX_QUERY)
        self.write_metadata(metadata)


    def create_indices(self):
        with closing(self.connect()) as conn,\
             closing(conn.cursor()) as cursor:
            cursor.execute(CREATE_READS_ID_INDEX_QUERY)
            cursor.execute(CREATE_SIGNAL_DATA_INDEX_QUERY)
            cursor.execute(CREATE_TRANSCRIPTS_ID_INDEX_QUERY)


    def write_metadata(self, metadata):
        with closing(self.connect()) as conn,\
             closing(conn.cursor()) as cursor:
            for key, value in metadata.items():
                cursor.execute("INSERT INTO metadata (key, value) VALUES (?, ?)",
                               (key, str(value)))


    def merge_in_databases(self, databases, merge_reads=True):
        with closing(self.connect()) as conn,\
             closing(conn.cursor()) as cursor:
            for other_db in databases:
                cursor.execute(f"ATTACH '{other_db}' as tmp")
                cursor.execute("BEGIN")
                cursor.execute("INSERT INTO transcripts SELECT * FROM tmp.transcripts")
                if merge_reads:
                    cursor.execute("INSERT INTO reads SELECT * FROM tmp.reads")
                cursor.execute("INSERT INTO signal_data SELECT * FROM tmp.signal_data")
                conn.commit()
                cursor.execute("DETACH DATABASE tmp")
                # Delete the other db
                Path(other_db).unlink()


    def get_references_with_data(self) -> dict[str, int]:
        with closing(self.connect()) as conn,\
             closing(conn.cursor()) as cursor:
            query = """
            SELECT t.name, COUNT(sd.read_id)
            FROM signal_data sd
            INNER JOIN transcripts t ON sd.transcript_id = t.id
            GROUP BY t.id
            """
            return {row[0]: row[1]
                    for row in cursor.execute(query).fetchall()}


    def get_references(self):
        with closing(self.connect()) as conn,\
             closing(conn.cursor()) as cursor:
            query = """
            SELECT DISTINCT name, id
            FROM transcripts
            """
            return {TranscriptRow(row[0], row[1])
                    for row in cursor.execute(query).fetchall()}


    @staticmethod
    def write_signal_data_rows(connection, rows):
        with closing(connection.cursor()) as cursor:
            cursor.execute("begin")
            cursor.executemany(INSERT_SIGNAL_DATA_QUERY, rows)
            cursor.execute("commit")


    @staticmethod
    def get_signal_data(connection,
                        transcript_name: str,
                        max_invalid_ratio: float):
        """
        Get signal data for a transcript.
        Will return reads with lower invalid ratio first.

        Parameters
        ----------
        connection : sqlite3.connection
            Connection to the database.
        transcript_name : str
            The name of the transcript (transcript reference).
        max_invalid_ratio : float
            Will return only reads with invalid ratio smaller
            than the given number.

        Returns
        -------
        list[tuple[bytearray, bytearray]]
            List of tuples (intensity, dwell), where
            intensity and dwell are binary arrays
            than need to be decoded to numpy arrays.
        """
        with closing(connection.cursor()) as cursor:
            return cursor.execute(GET_SIGNAL_DATA_FOR_TRANSCRIPT_QUERY,
                                  (transcript_name, max_invalid_ratio)).fetchall()




import multiprocessing
import os
import queue
import sqlite3
import sys
import time
import traceback

from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import as_completed
from pprint import pformat

import numpy as np
import pysam
import torch

from jaxtyping import Float, Int
from loguru import logger
from pyfaidx import Fasta

from nanocompore.common import INTENSITY_POS
from nanocompore.common import DWELL_POS
from nanocompore.common import MOTOR_DWELL_POS
from nanocompore.common import MEASUREMENTS_TYPE
from nanocompore.common import UNCALLED4
from nanocompore.common import TranscriptRow
from nanocompore.common import encode_kmer
from nanocompore.common import get_reads_invalid_ratio
from nanocompore.common import get_references_from_bam
from nanocompore.common import get_pos_kmer
from nanocompore.common import log_init_state
from nanocompore.comparisons import TranscriptComparator
from nanocompore.config import Config
from nanocompore.database import ResultsDB
from nanocompore.database import PreprocessingDB
from nanocompore.postprocessing import Postprocessor
from nanocompore.transcript import Transcript
from nanocompore.uncalled4 import Uncalled4


MAX_PROC_ITERATIONS = 500


class RunCmd(object):
    """
    RunCmd reads the resquiggled and preprocessed data
    and performs comparison between the samples of the two
    conditions for each position on every transcript to
    detect the presence of likely modifications.
    """

    def __init__(self, config: Config, random_state: int=42):
        """
        Initialise a `RunCmd` object and generates a white list of
        references with sufficient coverage for subsequent analysis.
        The retuned object can then be called to start the analysis.
        * config
            Config object containing all the parameters for the analysis.
        """
        logger.info("Checking and initialising RunCmd")

        # Save init options in dict for later
        log_init_state(loc=locals())

        self._config = config
        self._workers_with_device = self._get_workers_with_device()
        self._sample_labels = self._config.get_sample_labels()
        self._random_state = random_state


    def __call__(self):
        """
        The main method that will perform the comparative analysis.
        The function will get transcripts for processing from the input
        database, start worker processes to compare them and write the results
        in the output database, and finally will perform the postprocessing
        to export TSVs with the output.
        """
        logger.info("Starting data processing")
        logger.info(f"The input configuration is:\n{pformat(self._config._config)}")

        if self._config.get_progress():
            print("Getting the list of transcripts for processing...")

        db_path = ResultsDB(self._config, init_db=False).db_path
        db_exists = os.path.isfile(db_path)
        if db_exists and self._config.get_result_exists_strategy() == 'continue':
            db = ResultsDB(self._config, init_db=False)
            transcript_index_start = db.get_next_transcript_id()
            transcripts = self._get_transcripts_for_processing(initial_index=transcript_index_start)
            all_transcripts_num = len(transcripts)
            transcripts = self._filter_processed_transcripts(transcripts)
            already_processed = all_transcripts_num - len(transcripts)
            logger.info(f"Found a total of {all_transcripts_num}. " + \
                        f"{already_processed} have already been processed in previous runs. " + \
                        f"Will process only for the remaining {len(transcripts)}.")
        else:
            # Initialize the database
            ResultsDB(self._config, init_db=True)
            transcripts = self._get_transcripts_for_processing()
            logger.info(f"Found a total of {len(transcripts)} transcripts for processing.")

        task_queue = multiprocessing.JoinableQueue()
        for transcript_ref in transcripts:
            task_queue.put((transcript_ref, 0))
        logger.info("All transcripts have been sent to the queue")

        if self._config.get_progress():
            print("Starting to compare the conditions for all transcripts...")
        logger.info(f"Starting {len(self._workers_with_device)} worker processes.")
        db_lock = multiprocessing.Lock()
        sync_lock = multiprocessing.Lock()
        manager = multiprocessing.Manager()
        num_finished = manager.Value('i', 0)

        if self._config.get_resquiggler() == UNCALLED4:
            worker_class = Uncalled4Worker
        else:
            worker_class = GenericWorker
        workers = {worker: worker_class(worker,
                                        task_queue,
                                        db_lock,
                                        sync_lock,
                                        num_finished,
                                        device,
                                        self._config)
                   for worker, device in self._workers_with_device.items()}

        for worker in workers.values():
            worker.start()

        # helper variables for the progress bar
        max_len = 0
        start_time = time.time()

        # Wait for all workers to finish while printing
        # the progress bar if it was requested.
        # We use this slightly weird pattern of waiting
        # for processes instead of a standard loop with
        # unbounded join, because otherwise we'll have
        # to start a separate thread for the progress bar
        # monitoring and that doesn't work well with the
        # multiprocess Manager.
        while len(workers) > 0:
            workers_to_remove = []
            for worker_id, worker in workers.items():
                # Try to join the worker with 1s timeout
                worker.join(1)
                # If the worker has completed we remove it from
                # the worker pool.
                if worker.exitcode is not None:
                    if worker.exitcode == 0:
                        # If we detect a worker that has terminated without an error
                        # it may mean that there were no more jobs in the queue an
                        # the worker closed itself down, or it can be that the worker
                        # did its maximum allowed iterations. If we're in the second
                        # case we want to restart the worker if there are more tasks
                        # to do.
                        with sync_lock:
                            if not task_queue.empty():
                                device = self._workers_with_device[worker_id]
                                logger.info(f"Restarting worker {worker_id} on device {device}.")
                                new_worker = worker_class(worker_id,
                                                          task_queue,
                                                          db_lock,
                                                          sync_lock,
                                                          num_finished,
                                                          device,
                                                          self._config)
                                new_worker.start()
                                workers[worker_id] = new_worker
                            else:
                                workers_to_remove.append(worker_id)
                    else:
                        logger.error(f"ERROR: Worker {worker_id} encountered an error "
                                     f"(exitcode: {worker.exitcode}). "
                                     "Will terminate all other workers and stop.")
                        for child in multiprocessing.active_children():
                            child.terminate()
                        sys.exit(1)
                if self._config.get_progress():
                    with sync_lock:
                        num_done = num_finished.value
                    self._print_progress_bar(num_done, len(transcripts), max_len, start_time)
            for worker_id in workers_to_remove:
                del workers[worker_id]

        if self._config.get_progress():
            print("\nAll transcripts have been processed.")
            print("Starting postprocessing...")

        postprocessor = Postprocessor(self._config)
        postprocessor()

        logger.info(f"Done.")
        if self._config.get_progress():
            print("Done.")


    def _get_transcripts_for_processing(self, initial_index=1) -> set[TranscriptRow]:
        min_coverage = self._config.get_min_coverage()
        # References is a dict of the form
        # ref_id => [cond0_counts, cond1_counts]
        references = defaultdict(lambda: [0, 0])
        with ThreadPoolExecutor(max_workers=4) as executor:
            # We prepare functions to get data concurrently from
            # dbs or bams depending on the resquigglers.
            if self._config.get_resquiggler() == UNCALLED4:
                fn = lambda sample_def: executor.submit(get_references_from_bam,
                                                        sample_def['bam'])
            else:
                fn = lambda sample_def: executor.submit(
                        PreprocessingDB(sample_def['db']).get_references_with_data)

            futures = {fn(sample_def): cond
                       for cond, condition_def in self._config.get_data().items()
                       for sample, sample_def in condition_def.items()}

            cond_to_id = self._config.get_condition_ids()
            for future in as_completed(futures):
                condition = futures[future]
                condition_id = cond_to_id[condition]
                for ref_id, count in future.result().items():
                    references[ref_id][condition_id] += count

        # Get only refs that pass the min_coverage
        # requirements for both conditions.
        filtered_refs = {ref_id
                         for ref_id, (cond0_count, cond1_count) in references.items()
                         if cond0_count >= min_coverage and cond1_count >= min_coverage}

        # If selected_refs is set, we read the file
        # and keep only the references found within it.
        if self._config.get_selected_refs():
            with open(self._config.get_selected_refs()) as f:
                selected_refs = {ref.strip() for ref in f.readlines()}
                filtered_refs = {ref
                                 for ref in filtered_refs
                                 if ref in selected_refs}

        # If ignored_refs is set, we read the file
        # and discard the references found within it.
        if self._config.get_ignored_refs():
            with open(self._config.get_ignored_refs()) as f:
                ignored_refs = {ref.strip() for ref in f.readlines()}
                filtered_refs = {ref
                                 for ref in filtered_refs
                                 if ref not in ignored_refs}

        return {TranscriptRow(ref_id, i)
                for ref_id, i in zip(filtered_refs,
                                     range(initial_index,
                                           initial_index + len(filtered_refs)))}


    def _filter_processed_transcripts(self, transcripts: set[TranscriptRow]) -> set[TranscriptRow]:
        db = ResultsDB(self._config, init_db=False)
        logger.info(f'Preprocessing db: {db.db_path}')
        existing_transcripts = set(db.get_transcripts())
        return {tx for tx in transcripts if tx.ref_id not in existing_transcripts}


    def _get_workers_with_device(self):
        devices = self._config.get_devices()
        result = {}
        worker = 0
        for device, procs in devices.items():
            for _ in range(procs):
                result[worker] = device
                worker += 1
        return result


    def _print_progress_bar(self,
                            num_done: int,
                            total_transcripts: int,
                            max_len: int,
                            start_time: float):
        progress_bar_len = 30
        perc_done = num_done / total_transcripts
        done_chars = int(perc_done*progress_bar_len)
        done_bar = '▮' * done_chars
        remaining_bar = '-' * (progress_bar_len - done_chars)
        running_time = time.time() - start_time
        if num_done == 0:
            eta = 'ETA: N/A'
        else:
            total_expected = running_time * total_transcripts / num_done
            remaining_sec = total_expected - running_time
            hours, minutes, _ = self._format_time(remaining_sec)
            eta = f'ETA: {hours:02d}h:{minutes:02d}m remaining'
        hours, minutes, seconds = self._format_time(running_time)
        running = f'⏱  {hours:02d}h:{minutes:02d}m:{seconds:02d}s'
        speed = num_done / (running_time / 60)
        bar = f'{running} |{done_bar}{remaining_bar}| ' + \
              f'{num_done:,}/{total_transcripts:,} ' + \
              f'({perc_done*100:3.1f}%) [{speed:.1f}tx/m] | {eta}'
        if len(bar) > max_len:
            max_len = len(bar)
        if len(bar) < max_len:
            bar = f'{bar}{(max_len - len(bar)) * " "}'
        print(bar, end='\r', flush=True)
        return max_len


    def _format_time(self, seconds: float):
        hours, remainder = divmod(seconds, 3600)
        minutes, seconds = divmod(remainder, 60)
        return int(hours), int(minutes), int(seconds)


class Worker(multiprocessing.Process):
    def __init__(self,
                 worker_id,
                 task_queue,
                 db_lock,
                 sync_lock,
                 num_finished,
                 device,
                 config):
        """
        Abstract parent for different type of worker processes.
        Different workers may read data from different source,
        but then should process it in the same way.

        The worker will continuously take tasks (transcripts)
        from the task queue, process them and write the results
        to the database. Will quit when the task queue is empty.

        Parameters
        ----------
        worker_id : int
            Id of the worker
        task_queue : multiprocessing.JoinableQueue
            Queue with tasks for processing.
        db_lock : multiprocessing.Lock
            Lock to prevent parallel writing to the SQLite DB.
        sync_lock : multiprocessing.Lock
            Lock for synchronisation.
        num_finished : multiprocessing.managers.SyncManager
            Number of processed transcripts.
        device : str
            Which device to use for computation (e.g. cpu or gpu).
        config : nanocompore.Config
            The configuration object.
        """
        super().__init__()
        self._id = worker_id
        self._task_queue = task_queue
        self._db_lock = db_lock
        self._sync_lock = sync_lock
        self._num_finished = num_finished
        self._device = device
        # Note: we're deviating from our convention
        # to use self._config everywhere for the
        # Config object, because _config is used
        # by Process.
        self._conf = config
        self._kit = self._conf.get_kit()
        self.log("info", f"Starting on pid {os.getpid()}.")


    def run(self):
        # We run a second post __init__
        # setup method in order to ensure
        # that it is executed within the
        # worker's process and not on the
        # parent process. This allows us
        # to use spawn() to create the worker
        # process and not only fork().
        self.setup()

        # When Python grows the heap to find space
        # for memory allocations, after the objects
        # are deleted and freed, the memory may not
        # be returned back to the OS. Hence, the
        # reserved memory stays at the high-water
        # mark level that was previously taken.
        # Since each worker would stumble at an
        # unusally large transcript (in terms of
        # the amount of data) from time to time,
        # the heap of the worker may grow, but it
        # won't shrink after that, leading to
        # a lot of unused head size. When this happens
        # for many workers, the OS or job scheduler
        # may kill the process because it used too
        # much memory.
        # To ameliorate this issue, we can set a
        # fixed number of iterations for the worker
        # and let the main orchestrator process
        # spawn a new worker when it detects the
        # termination.
        for _ in range(MAX_PROC_ITERATIONS):
            # It's weird that we have to use a lock here,
            # but for some reason calling get in parallel from
            # multiple processes can cause a queue.Empty
            # exception to be raised even for non-empty queues...
            # https://github.com/python/cpython/issues/87302
            with self._sync_lock:
                try:
                    self.log("info",
                             f"Getting a task from a queue with size {self._task_queue.qsize()}")
                    msg = self._task_queue.get(block=False)
                except queue.Empty:
                    self.log("info", "Cannot find more tasks in the queue and will terminate.")
                    break

            transcript_ref, retries = msg
            if retries > 2:
                self.log("error",
                         f"Transcript {transcript_ref.ref_id} was retried too many times. "
                         "Won't retry again")
                continue
            self.log("info", f"Starting to process {transcript_ref.ref_id}")
            transcript = Transcript(
                    id=transcript_ref.id,
                    ref_id=transcript_ref.ref_id,
                    ref_seq=str(self._fasta_fh[transcript_ref.ref_id]))

            try:

                start_time = time.time()
                data, samples, conditions = self._read_data(transcript)
                self.log("debug",
                         f"Read data for transcript {transcript.name} in "
                         f"{time.time() - start_time}.")
                # If we don't have valid reads, just continue
                if data.shape[1] == 0:
                    with self._db_lock:
                        self.log("info",
                                 f"Not enough valid reads for {transcript.name}. "
                                 "Saving the transcript as processed.")
                        self._db_manager.save_transcript(transcript)
                    continue

                prepared_data = self._prepare_data(data, samples, conditions)
                data, samples, conditions, positions = prepared_data
                min_cov = self._conf.get_min_coverage()
                data, positions = self._filter_low_cov_positions(data, positions, conditions, min_cov)
                max_reads = self._conf.get_downsample_high_coverage()
                data, samples, conditions = self._downsample(
                        data, samples, conditions, max_reads)

                transcript, results = self._comparator.compare_transcript(
                        transcript,
                        data,
                        samples,
                        conditions,
                        positions,
                        self._device)

                if results is None:
                    with self._db_lock:
                        self.log("info",
                                 f"No test results for {transcript.name}. "
                                 "Saving the transcript as processed.")
                        self._db_manager.save_transcript(transcript)
                    continue

                seq = transcript.seq
                kmers = results.pos.apply(lambda pos: get_pos_kmer(pos, seq, self._kit))
                results['kmer'] = kmers.apply(encode_kmer)
                with self._db_lock:
                    self.log("info", f"Saving the results for {transcript.name}.")
                    self._db_manager.save_test_results(transcript, results)
            except torch.OutOfMemoryError:
                task = TranscriptRow(transcript.name, transcript.id)
                self.log("info",
                         f"Returning {transcript_ref.ref_id} to the queue for processing it "
                         "later because it failed with an OutOfMemory")
                with self._sync_lock:
                    self._task_queue.put((task, retries + 1))
            except Exception as e:
                if 'out of memory' in str(e):
                    task = TranscriptRow(transcript.name, transcript.id)
                    logger.info(f"Worker {self._id} returns {transcript_ref.ref_id} to " + \
                                 "the queue for later processing because it failed with OutOfMemory.")
                    with self._sync_lock:
                        self._task_queue.put((task, retries + 1))
                else:
                    msg = traceback.format_exc()
                    self.log("error",
                             f"Encountered an error for {transcript_ref.ref_id}: {msg}")
            finally:
                self.log("info", f"Finishing a task: {transcript_ref.ref_id}.")
                with self._sync_lock:
                    self._num_finished.value += 1
                    self._task_queue.task_done()
        else:
            self.log("info", "Completed a batch of transcripts. Will restart now.")


    def log(self, level, msg):
        # We use opt(depth=1)" in order to write the line
        # number from which the log function is called.
        if level.upper() == "ERROR":
            logger.opt(depth=1).error(f"[Worker {self._id}] {msg}")
        elif level.upper() == "WARNING":
            logger.opt(depth=1).warning(f"[Worker {self._id}] {msg}")
        elif level.upper() == "INFO":
            logger.opt(depth=1).info(f"[Worker {self._id}] {msg}")
        elif level.upper() == "DEBUG":
            logger.opt(depth=1).debug(f"[Worker {self._id}] {msg}")
        elif level.upper() == "TRACE":
            logger.opt(depth=1).trace(f"[Worker {self._id}] {msg}")
        else:
            raise ValueError(f"Unrecognized log level: {level}")


    def _filter_low_cov_positions(self, data, positions, conditions, min_cov):
        cov_cond_0 = torch.sum(~data[:, conditions == 0, INTENSITY_POS].isnan(), dim=1)
        cov_cond_1 = torch.sum(~data[:, conditions == 1, INTENSITY_POS].isnan(), dim=1)
        valid_positions = torch.logical_and(cov_cond_0 >= min_cov, cov_cond_1 >= min_cov)
        return data[valid_positions], positions[valid_positions]


    def _downsample(self, data, samples, conditions, max_reads):
        if data.shape[1] > max_reads:
            read_valid_positions = (~data.isnan().any(2)).sum(0)
            read_order = read_valid_positions.argsort(descending=True)
            selected = torch.full((read_order.shape[0],), False)
            for cond in [0, 1]:
                cond_selected = read_order[conditions[read_order] == cond][:max_reads]
                selected[cond_selected] = True
            data = data[:, selected, :]
            samples = samples[selected]
            conditions = conditions[selected]
        return data, samples, conditions


    def _prepare_data(self,
                      data: Float[np.ndarray, "positions reads vars"],
                      sample_ids: Int[np.ndarray, "reads"],
                      condition_ids: Int[np.ndarray, "reads"],
    ) -> tuple[Float[torch.Tensor, "positions reads vars"],
               Float[torch.Tensor, "reads"],
               Float[torch.Tensor, "reads"],
               Int[torch.Tensor, "positions"]]:
        """
        Prepare data for comparisons.
        This will convert the data to pytorch tensor
        and add the motor dwell time if necessary.

        Parameters
        ----------
        data : Float[np.ndarray, "positions reads vars"]
            Measurements tensor with shape (Positions, Reads, Vars)
        sample_ids : Int[np.ndarray, "reads"]
            1D array of sample ids with size R
        condition_ids : Int[np.ndarray, "reads"]
            1D array of condition ids with size R

        Returns
        -------
        tuple[Float[torch.Tensor, "positions reads vars"],
              Float[torch.Tensor, "reads"],
              Float[torch.Tensor, "reads"],
              Int[torch.Tensor, "positions"]]
            Tuple with:
            - Tensor with shape (Positions, Reads, Vars) containing the
              measurement data.
            - 1D tensor of sample ids with size R
            - 1D tensor of condition ids with size R
            - 1D tensor of reference positions with size P
        """
        # We use the log10 of the dwell time. This
        # significantly improves the accuracy of the detection.
        # We add a small epsilon number to make sure there's no
        # division by zero when calculating the logarithm.
        # As all data is shifted with the same amount, this
        # shouldn't have any effect on the modification detection.
        data[:, :, DWELL_POS] = np.log10(data[:, :, DWELL_POS] + 1e-10)

        motor_offset = self._conf.get_motor_dwell_offset()
        num_positions = data.shape[0]
        if self._conf.get_motor_dwell_offset() > 0:
            data = np.pad(data,
                          ((0, 0), (0, 0), (0, 1)),
                          mode='constant',
                          constant_values=np.nan)
            end = num_positions - motor_offset
            if end > 0:
                data[:end, :, MOTOR_DWELL_POS] = data[motor_offset:, :, DWELL_POS]

        # Keep only positions for which we have at least one read.
        valid_positions = np.any(~np.isnan(data[:, :, INTENSITY_POS]), axis=1)
        tensor = torch.tensor(data[valid_positions, :, :],
                              dtype=torch.float32,
                              device=self._device)
        samples = torch.tensor(sample_ids,
                               dtype=torch.int16,
                               device=self._device)
        conditions = torch.tensor(condition_ids,
                                  dtype=torch.int16,
                                  device=self._device)
        positions = torch.arange(data.shape[0],
                                 dtype=torch.int16,
                                 device=self._device)[valid_positions]

        return (tensor, samples, conditions, positions)


    def setup(self):
        """
        Setup the worker.

        This is used to add non-serializable properties
        to the instance.

        This is separate from __init__ to ensure
        that it is executed in the child process.
        This allows using both fork() and spawn()
        to create a worker process.
        """
        self._db_manager = ResultsDB(self._conf, init_db=False)
        self._fasta_fh = Fasta(self._conf.get_fasta_ref())
        self._comparator = TranscriptComparator(config=self._conf, worker=self)


    def _read_data(
        self,
        transcript: Transcript
    ) -> tuple[Float[np.ndarray, "positions reads vars"],
               Float[np.ndarray, "reads"],
               Float[np.ndarray, "reads"]]:
        """
        Get read data for the given transcript.

        Parameters
        ----------
        transcript : Transcript
            Transcript for which to acqire data.

        Returns
        -------
        tuple[Float[np.ndarray, "positions reads vars"],
              Float[np.ndarray, "reads"],
              Float[np.ndarray, "reads"]]
            Tuple with (measurements tensor, sample ids, condition ids).
            - measurements tensor: shape (Positions, Reads, Vars)
            - sample ids: 1D array with int ids of the samples
            - condition ids: 1D array with int ids of the conditinos (0 or 1).
        """
        raise NotImplementedError("This method should be overriden by specific " + \
                                  "worker process implementations.")


class Uncalled4Worker(Worker):
    def setup(self):
        """
        Setup the GenericWorker.

        This is used to add non-serializable properties
        to the instance.

        This is separate from __init__ to ensure
        that it is executed in the child process.
        This allows using both fork() and spawn()
        to create a worker process.
        """
        super().setup()
        data_def = self._conf.get_sample_condition_bam_data()
        self._bams = {sample: pysam.AlignmentFile(bam, 'rb')
                      for sample, _, bam in data_def}


    def _read_data(
        self,
        transcript: Transcript
    ) -> tuple[Float[np.ndarray, "positions reads vars"],
               Float[np.ndarray, "reads"],
               Float[np.ndarray, "reads"]]:
        """
        Get read data for the given transcript.

        Parameters
        ----------
        transcript : Transcript
            Transcript for which to acqire data.

        Returns
        -------
        tuple[Float[np.ndarray, "positions reads vars"],
              Float[np.ndarray, "reads"],
              Float[np.ndarray, "reads"]]
            Tuple with (measurements tensor, sample ids, condition ids).
            - measurements tensor: shape (Positions, Reads, Vars)
            - sample ids: 1D array with int ids of the samples
            - condition ids: 1D array with int ids of the conditinos (0 or 1).
        """
        uncalled4 = Uncalled4(transcript.name,
                              len(transcript.seq),
                              self._bams,
                              self._conf.get_kit())
        data, reads, samples = uncalled4.get_data()

        order = np.argsort(reads)
        data = data[:, order, :]
        samples = samples[order]

        # convert sample labels to sample int ids
        sample_id_mapper = np.vectorize(self._conf.get_sample_ids().get)
        sample_ids = sample_id_mapper(samples)

        # convert sample labels to condition int ids
        samp_to_cond_mapper = np.vectorize(self._conf.sample_to_condition().get)
        condition_id_mapper = np.vectorize(self._conf.get_condition_ids().get)
        condition_ids = condition_id_mapper(samp_to_cond_mapper(samples))

        invalid_ratios = get_reads_invalid_ratio(data[:, :, INTENSITY_POS])
        valid_reads = invalid_ratios < self._conf.get_max_invalid_kmers_freq()

        self.log("debug",
                 f"Filtering uncalled4 reads in read_data: "
                 f"valid {valid_reads.sum()} out of {valid_reads.shape[0]}")

        return data[:, valid_reads], sample_ids[valid_reads], condition_ids[valid_reads]


    def close(self):
        for _, bam in self._bams:
            bam.close()
        super().close()


class GenericWorker(Worker):
    def setup(self):
        """
        Setup the GenericWorker.

        This is used to add non-serializable properties
        to the instance.

        This is separate from __init__ to ensure
        that it is executed in the child process.
        This allows using both fork and spawn to
        create a worker process.
        """
        super().setup()
        data_def = self._conf.get_sample_condition_db_data()
        self._dbs = {sample: sqlite3.connect(db, check_same_thread=False)
                     for sample, _, db in data_def}


    def _read_data(
        self,
        transcript: Transcript
    ) -> tuple[Float[np.ndarray, "positions reads vars"],
               Float[np.ndarray, "reads"],
               Float[np.ndarray, "reads"]]:
        """
        Get read data for the given transcript.

        Parameters
        ----------
        transcript : Transcript
            Transcript for which to acqire data.

        Returns
        -------
        tuple[Float[np.ndarray, "positions reads vars"],
              Float[np.ndarray, "reads"],
              Float[np.ndarray, "reads"]]
            Tuple with (measurements tensor, sample ids, condition ids).
            - measurements tensor: shape (Positions, Reads, Vars)
            - sample ids: 1D array with int ids of the samples
            - condition ids: 1D array with int ids of the conditinos (0 or 1).
        """
        intensities = {}
        dwells = {}
        sample_counts = {}
        n_samples = len(self._dbs)
        # Here we will use multi-threaded reading from
        # the input databases (one thread per sample)
        # and then we'll combine the reads into one
        # large tensor. However, we take care to
        # ensure that the data is read in the same
        # order each time in order to make the
        # GMM fitting deterministic.
        with ThreadPoolExecutor(max_workers=n_samples) as executor:
            futures = [executor.submit(get_transcript_signals_from_db,
                                       (self._dbs[sample],
                                        transcript.name,
                                        self._conf.get_max_invalid_kmers_freq(),
                                        sample,
                                        condition))
                       for sample, condition, _ in self._conf.get_sample_condition_db_data()]
            for future in as_completed(futures):
                signal_data, sample, _ = future.result()
                sample_counts[sample] = len(signal_data)
                if len(signal_data) == 0:
                    continue
                # intensity will have shape (Reads, Positions)
                intensity = np.array([np.frombuffer(intensity, dtype=MEASUREMENTS_TYPE)
                                      for intensity, _ in signal_data])
                intensities[sample] = intensity
                # dwell will have shape (Reads, Positions)
                dwell = np.array([np.frombuffer(dwell, dtype=MEASUREMENTS_TYPE)
                                  for _, dwell in signal_data])
                dwells[sample] = dwell

        if all([c == 0 for c in sample_counts.values()]):
            return np.empty((0, 0, 0)), np.array([]), np.array([])

        ordered_intensities = [intensities[s]
                               for s in self._conf.get_sample_labels()
                               if s in intensities]
        ordered_dwells = [dwells[s]
                          for s in self._conf.get_sample_labels()
                          if s in dwells]

        # The tensor has shape: (Positions, Reads, Vars)
        # Note, intensity and dwell have shape (Reads, Positions)
        # so we transpose them before stacking them.
        tensor = np.stack([np.concatenate(ordered_intensities).T,
                           np.concatenate(ordered_dwells).T],
                          axis=2)

        sample_ids = np.concatenate([np.full(sample_counts[samp], sid)
                                     for samp, sid in self._conf.get_sample_ids().items()])
        samp_to_cond = self._conf.sample_to_condition()
        cond_label_to_id = self._conf.get_condition_ids()
        condition_ids = np.concatenate([np.full(sample_counts[samp],
                                                cond_label_to_id[samp_to_cond[samp]])
                                        for samp in self._conf.get_sample_ids().keys()])

        return tensor, sample_ids, condition_ids


    def close(self):
        for _, db in self._dbs:
            db.close()
        super().close()


def get_transcript_signals_from_db(args):
    (connection,
     transcript_name,
     max_invalid_ratio,
     sample,
     condition) = args
    # Note: we use config.get_downsample_high_coverage to
    # set a limit on the number of reads per condition.
    # However, it will be pointless to take more reads
    # for a given sample than we wull use for the condition.
    # For this reason, we limit the per-sample reads to
    # the same number, to avoid reading and processing
    # large amount of data unnecessarily. When combining
    # the data from different samples of the same conditions
    # some of those reads would be discarded potentially.
    signal_data = PreprocessingDB.get_signal_data(connection,
                                                  transcript_name,
                                                  max_invalid_ratio)
    return signal_data, sample, condition




"""
Reads an eventalign tsv file produced by
Nanopolish or f5c and creates a collapsed
version of the file that merges together
events for the same position.
The output is written to a compact SQLite
database file.
"""
import multiprocessing as mp
import os
import sqlite3
import statistics
import sys
import threading
import traceback
import uuid

from jaxtyping import Float, Bool
from contextlib import closing
from pathlib import Path

import numpy as np

from loguru import logger
from pyfaidx import Fasta

from nanocompore.database import CREATE_SIGNAL_DATA_TABLE_QUERY
from nanocompore.database import CREATE_READS_TABLE_QUERY
from nanocompore.database import CREATE_TRANSCRIPTS_TABLE_QUERY
from nanocompore.database import INSERT_READS_QUERY
from nanocompore.database import INSERT_SIGNAL_DATA_QUERY
from nanocompore.database import INSERT_TRANSCRIPTS_QUERY
from nanocompore.database import CREATE_READS_ID_INDEX_QUERY
from nanocompore.database import CREATE_SIGNAL_DATA_INDEX_QUERY
from nanocompore.common import READ_ID_TYPE
from nanocompore.common import MEASUREMENTS_TYPE
from nanocompore.common import monitor_workers


class Kmer:
    def __init__(self, pos, measurements, dwell, valid, kmer):
        self.pos = pos
        self.measurements = measurements
        self.dwell = dwell
        self.valid = valid
        self.kmer = kmer


class EventalignCollapser:
    """
    Parses an eventalign TSV (either given as a
    file or piped to the STDIN of the current
    process), collapses the events for each
    transcript position and stores the relevant
    data to an intermediary SQLite database.
    """
    def __init__(self, file, fasta_ref, output, nthreads, tmp_dir):
        if nthreads < 2:
            raise ValueError("At least 2 threads required.")

        self._file = file
        self._fasta_ref = fasta_ref
        self._ref_lens = {}
        self._nthreads = nthreads
        self._output = output
        self._tmp_dir = tmp_dir


    def __call__(self):
        manager = mp.Manager()
        # nthreads is num_workers + 1, so using this
        # number as a size of the queue means that
        # if the producer is faster than the workers
        # then there'll always be one task ready
        # for processing once a worker gets free.
        # Note: using limited queue is important
        # for the streaming (pipe) case to prevent
        # flooding the file system with big files.
        task_queue = manager.Queue(maxsize=self._nthreads)
        lock = manager.Lock()
        transcript_id = manager.Value('i', 1)
        reads_offset = manager.Value('i', 0)

        num_workers = self._nthreads - 1

        workers = [mp.Process(target=self._worker,
                              args=(i,
                                    task_queue,
                                    lock,
                                    transcript_id,
                                    reads_offset))
                   for i in range(num_workers)]

        for worker in workers:
            worker.start()

        worker_monitor = threading.Thread(target=monitor_workers, args=(workers, ))
        worker_monitor.start()

        if self._file:
            self._file_indexer(task_queue)
        else:
            self._stream_indexer(task_queue)

        # add poison pills to kill the workers
        for _ in range(num_workers):
            task_queue.put(None)

        for worker in workers:
            worker.join()
        worker_monitor.join()

        logger.info("Merging tmp databases.")

        # Make sure there's no old existing database file.
        # If a previous run was abruptly stopped the database
        # may be corrupted, thus breaking the process when
        # we try to open it.
        db_path = Path(self._output)
        if db_path.is_file():
            db_path.unlink()
        self._create_tables(self._output)

        with closing(self._get_db(self._output)) as conn,\
             closing(conn.cursor()) as cursor:
            for i in range(num_workers):
                tmp_db = self._output + f".{i}"
                cursor.execute(f"ATTACH '{tmp_db}' as tmp")
                cursor.execute("BEGIN")
                cursor.execute("INSERT INTO transcripts SELECT * FROM tmp.transcripts")
                cursor.execute("INSERT INTO reads SELECT * FROM tmp.reads")
                cursor.execute("INSERT INTO signal_data SELECT * FROM tmp.signal_data")
                conn.commit()
                cursor.execute("DETACH DATABASE tmp")
                # Delete the tmp db
                Path(tmp_db).unlink()

        logger.info("Creating database indices.")
        self._create_db_indices()


    def _file_indexer(self, task_queue):
        logger.info("The indexer is starting to parse the input eventalign tsv.")
        with open(self._file, 'r') as file:
            # ignore the header
            file.readline()

            prev_ref_id = None
            ref_start = file.tell()
            current_line_pos = None
            while True:
                current_line_pos = file.tell()
                line = file.readline()
                if not line:
                    break

                cols = line.split('\t')
                ref_id = cols[0]

                if not prev_ref_id:
                    prev_ref_id = ref_id

                if ref_id != prev_ref_id:
                    task_queue.put((prev_ref_id, ref_start, current_line_pos))
                    ref_start = current_line_pos
                    prev_ref_id = ref_id
            if prev_ref_id:
                task_queue.put((prev_ref_id, ref_start, current_line_pos))


    def _stream_indexer(self, task_queue):
        # ignore the header
        sys.stdin.readline()

        tmp_file_name = os.path.join(self._tmp_dir,
                                     f'tmp_{str(uuid.uuid4())}')
        tmp_file = open(tmp_file_name, 'w')
        prev_ref_id = None
        while True:
            line = sys.stdin.readline()
            if not line:
                break

            ref_id, _ = line.split('\t', 1)

            if not prev_ref_id:
                prev_ref_id = ref_id

            if ref_id != prev_ref_id:
                task_queue.put((prev_ref_id, tmp_file_name))

                tmp_file.close()
                tmp_file_name = os.path.join(self._tmp_dir,
                                             f'tmp_{str(uuid.uuid4())}')
                tmp_file = open(tmp_file_name, 'w')

                prev_ref_id = ref_id

            tmp_file.write(line)

        if prev_ref_id:
            task_queue.put((prev_ref_id, tmp_file_name))
            tmp_file.close()


    def _read_data(self, file, fstart, fend) -> list[tuple[str, list[Kmer]]]:
        # The format is:
        # [
        #  ('read_id', [Kmer_pos1, Kmer_pos2, Kmer_pos3...], # read 1
        #  ('read_id', [Kmer_pos1, Kmer_pos2, Kmer_pos3...], # read 2
        # ]
        data = []

        file.seek(fstart)

        logger.info(f"Starting to read lines in range [{fstart}, {fend})")

        prev_read = None
        prev_pos = None
        kmer_data = None
        while True:
            line = file.readline()
            if file.tell() > fend or not line:
                break

            cols = line.split('\t')
            # The position in the eventalign is
            # the 0-based index of the initial
            # base in the k-mer.
            pos = int(cols[1])
            # reference kmer
            kmer = cols[2]
            read = cols[3]
            event_measurements = [float(v) for v in cols[13].split(',')]
            event_dwell = float(cols[8])
            model_kmer = cols[9]

            if read != prev_read:
                # If we have a previous kmer, we can finalize it
                # and calculate the median to store only the result
                # and freeing the memory used for all the sampled
                # measurements.
                if kmer_data:
                    # Note: the choice of using statistics' median
                    # instead of numpy's is not arbitrary. It seems
                    # that statistics.median is faster for small
                    # arrays, which is the expected case here.
                    kmer_data.median_intensity = statistics.median(kmer_data.measurements)
                    del kmer_data.measurements
                kmer_data = Kmer(pos,
                                 event_measurements,
                                 event_dwell,
                                 kmer == model_kmer,
                                 kmer)
                data.append((read, [kmer_data]))
                prev_read = read
                prev_pos = pos
            elif pos != prev_pos:
                if kmer_data:
                    kmer_data.median_intensity = statistics.median(kmer_data.measurements)
                    del kmer_data.measurements
                kmer_data = Kmer(pos,
                                 event_measurements,
                                 event_dwell,
                                 kmer == model_kmer,
                                 kmer)
                data[-1][1].append(kmer_data)
                prev_pos = pos
            else:
                # Get last read, read_data (second element in tuple), last kmer
                kmer_data = data[-1][1][-1]
                kmer_data.measurements.extend(event_measurements)
                kmer_data.dwell += event_dwell
                if kmer_data.valid:
                    kmer_data.valid = kmer == model_kmer
        if kmer_data:
            kmer_data.median_intensity = statistics.median(kmer_data.measurements)
            del kmer_data.measurements

        return data


    def _worker(self, idx, queue, lock, current_transcript_id, current_reads_offset):
        fasta = Fasta(self._fasta_ref)

        tmp_db = self._output + f".{idx}"
        # Make sure there's no old existing database file.
        # If a previous run was abruptly stopped the database
        # may be corrupted, thus breaking the process when
        # we try to open it.
        db_path = Path(tmp_db)
        if db_path.is_file():
            db_path.unlink()
        self._create_tables(tmp_db)

        if self._file:
            file = open(self._file, 'r')

        with closing(self._get_db(tmp_db)) as conn,\
             closing(conn.cursor()) as cursor:
            while True:
                task = queue.get()
                if not task:
                    break

                # If we have an eventalign file we can
                # read only the relevant parts with seek.
                # Else, if the eventalign data is streamed
                # through a pipe, we expect the parser to
                # create tmp files for the workers to process.
                if self._file:
                    ref_id, fstart, fend = task
                else:
                    ref_id, filepath = task
                    file = open(filepath, 'r')
                    fstart = 0
                    fend = np.inf

                try:
                    logger.info(f"Starting to read and process data for ref {ref_id}.")
                    data = self._read_data(file, fstart, fend)

                    ref_len = len(str(fasta[ref_id]))
                    intensity, dwell, reads_invalid_ratio = self._process_ref(ref_id, data, ref_len)

                    all_reads = list(reads_invalid_ratio.keys())

                    lock.acquire()
                    transcript_id = current_transcript_id.value
                    current_transcript_id.value += 1
                    reads_offset = current_reads_offset.value
                    current_reads_offset.value += len(all_reads)
                    lock.release()

                    read_ids = {read: i + reads_offset
                                for i, read in enumerate(all_reads)}

                    rows = self._process_rows_for_writing(transcript_id,
                                                          list(read_ids.values()),
                                                          intensity,
                                                          dwell)


                    reads_data = [(read, idx, reads_invalid_ratio[read])
                                  for read, idx in read_ids.items()]

                    cursor.execute('BEGIN')
                    cursor.execute(INSERT_TRANSCRIPTS_QUERY,
                                   (transcript_id, ref_id))
                    cursor.executemany(INSERT_READS_QUERY, reads_data)
                    cursor.executemany(INSERT_SIGNAL_DATA_QUERY, list(rows))
                    cursor.execute('COMMIT')
                    logger.info(f"Wrote transcript {transcript_id} {ref_id} to tmp database.")
                except Exception:
                    exception_msg = traceback.format_exc()
                    logger.error(f"Exception for ref {ref_id}: {exception_msg}")
                finally:
                    if not self._file:
                        file.close()
                        Path(filepath).unlink()


    def _process_ref(
        self,
        ref_id: str,
        data: list[tuple[str, list[Kmer]]],
        ref_len: int
    ) -> tuple[
            Float[np.ndarray, "reads positions"],
            Float[np.ndarray, "reads positions"],
            dict[str, float]]:
        nreads = len(data)

        intensity = np.empty((nreads, ref_len), dtype=MEASUREMENTS_TYPE)
        intensity.fill(np.nan)
        # We don't use sd for now
        # sd = np.empty((nreads, ref_len), dtype=MEASUREMENTS_TYPE)
        # sd.fill(np.nan)
        dwell = np.empty((nreads, ref_len), dtype=MEASUREMENTS_TYPE)
        dwell.fill(np.nan)
        valid = np.full((nreads, ref_len), False)
        read_ids = []
        starts = np.full((nreads,), np.iinfo(np.uint32).max, dtype=np.uint32)
        ends = np.full((nreads,), 0, dtype=np.uint32)

        for row, (read_id, read_data) in enumerate(data):
            read_ids.append(read_id)
            for kmer_data in read_data:
                pos = kmer_data.pos
                intensity[row, pos] = kmer_data.median_intensity
                dwell[row, pos] = kmer_data.dwell
                valid[row, pos] = kmer_data.valid
                starts[row] = min(starts[row], pos)
                ends[row] = max(ends[row], pos)

        # Calculate the ratio of invalid positions for all reads
        alignment_lengths = ends - starts + 1
        invalid_ratios = (alignment_lengths - valid.sum(axis=1))/alignment_lengths
        reads_invalid_ratio = dict(zip(read_ids, invalid_ratios))

        return intensity, dwell, reads_invalid_ratio


    def _process_rows_for_writing(self, transcript_id, read_ids, intensity, dwell):
        for i in range(intensity.shape[0]):
            yield (transcript_id, read_ids[i], intensity[i], dwell[i])


    def _get_db(self, path):
        conn = sqlite3.connect(path, isolation_level=None)
        conn.execute('PRAGMA synchronous = OFF')
        conn.execute('PRAGMA journal_mode = OFF')
        # Use largest possible page size of 65kb
        conn.execute('PRAGMA page_size = 65536')
        # Use cache of 2048 x page_size. That's 128MB of cache.
        conn.execute('PRAGMA cache_size = 2000')
        return conn


    def _create_tables(self, path):
        with closing(self._get_db(path)) as conn,\
             closing(conn.cursor()) as cursor:
            cursor.execute(CREATE_TRANSCRIPTS_TABLE_QUERY)
            cursor.execute(CREATE_READS_TABLE_QUERY)
            cursor.execute(CREATE_SIGNAL_DATA_TABLE_QUERY)


    def _create_db_indices(self):
        with closing(self._get_db(self._output)) as conn,\
             closing(conn.cursor()) as cursor:
            cursor.execute(CREATE_READS_ID_INDEX_QUERY)
            cursor.execute(CREATE_SIGNAL_DATA_INDEX_QUERY)




from nanocompore.common import get_references_from_bam
from nanocompore.config import Config
from tests.common import BASIC_CONFIG


def test_get_references_from_bams():
    config = Config(BASIC_CONFIG)
    references = get_references_from_bam(config.get_data()['kd']['kd1']['bam'])
    assert references == {'ENST00000674681.1|ENSG00000075624.17|OTTHUMG00000023268|-|ACTB-219|ACTB|2554|protein_coding|': 3,
                          'ENST00000642480.2|ENSG00000075624.17|OTTHUMG00000023268|OTTHUMT00000495153.1|ACTB-213|ACTB|2021|protein_coding|': 3}




import os

import torch


cwd = os.getcwd()

BASIC_CONFIG = {
    'data': {
        'kd': {
            'kd1': {
                'bam': os.path.join(cwd, 'tests/fixtures/kd1.bam'),
            },
            'kd2': {
                'bam': os.path.join(cwd, 'tests/fixtures/kd2.bam'),
            }
        },
        'wt': {
            'wt1': {
                'bam': os.path.join(cwd, 'tests/fixtures/wt1.bam'),
            },
            'wt2': {
                'bam': os.path.join(cwd, 'tests/fixtures/wt2.bam'),
            }
        }
    },
    'fasta': os.path.join(cwd, 'tests/fixtures/test_reference.fa'),
    'depleted_condition': 'kd',
    'kit': 'RNA002',
    'resquiggler': 'uncalled4',
}

BASIC_EVENTALIGN_CONFIG = {
    'data': {
        'kd': {
            'kd1': {
                'db': os.path.join(cwd, 'tests/fixtures/kd1_eventalign.db'),
            },
            'kd2': {
                'db': os.path.join(cwd, 'tests/fixtures/kd2_eventalign.db'),
            }
        },
        'wt': {
            'wt1': {
                'db': os.path.join(cwd, 'tests/fixtures/wt1_eventalign.db'),
            },
            'wt2': {
                'db': os.path.join(cwd, 'tests/fixtures/wt2_eventalign.db'),
            }
        }
    },
    'fasta': os.path.join(cwd, 'tests/fixtures/test_reference.fa'),
    'depleted_condition': 'kd',
    'kit': 'RNA002',
    'resquiggler': 'eventalign',
}

class MockWorker:
    def log(self, level, msg):
        pass


def naneq(a, b):
    """
    Compare two torch tensors for equality, handling nans.
    """
    if a.shape != b.shape:
        return False
    nans = a.isnan() & b.isnan()
    return torch.eq(a[~nans], b[~nans]).all().item()




import multiprocessing as mp
import os
import sqlite3
import threading

from collections import defaultdict
from concurrent.futures import ProcessPoolExecutor
from concurrent.futures import as_completed
from contextlib import closing
from pathlib import Path
from typing import Any

import pysam
import numpy as np

from jaxtyping import Float, Int
from loguru import logger
from pyfaidx import Fasta

from nanocompore.common import Kit
from nanocompore.common import READ_ID_TYPE
from nanocompore.common import get_reads_invalid_ratio
from nanocompore.common import get_references_from_bam
from nanocompore.common import monitor_workers
from nanocompore.database import INSERT_READS_QUERY
from nanocompore.database import INSERT_TRANSCRIPTS_QUERY
from nanocompore.database import PreprocessingDB
from nanocompore.remora_wrapper import Remora


class Preprocessor:
    """
    The preprocessor takes data coming from
    the chosen resquiggler (described in the
    configuration file used for running the
    experiment) and prepares it in a common
    format in a SQLite DB to be used in the
    subsequest analysis step.
    """
    def __init__(self, fasta_ref: str, output: str, threads: int):
        self._fasta_ref = fasta_ref
        self._output = output
        self._db = PreprocessingDB(output)
        self._db.setup(self._get_metadata())

        self._current_transcript_id = 1
        self._current_read_id = 1

        # The preprocessor will use the main process
        # and will spawn n - 1 worker processes in
        # order to make sure we use the number of
        # processes specified by the user in the config.
        self._worker_processes = threads - 1
        self._references = self._get_references()


    def __call__(self):
        manager = mp.Manager()
        task_queue = manager.Queue()
        lock = manager.Lock()
        current_transcript_id = manager.Value('i', 1)
        current_read_offset = manager.Value('i', 0)
        processed_transcripts = manager.Value('i', 0)

        logger.info(f"Starting {self._worker_processes} worker processes.")

        workers = [mp.Process(target=self._resquiggle,
                              args=(i,
                                    task_queue,
                                    lock,
                                    current_transcript_id,
                                    current_read_offset,
                                    processed_transcripts,
                                    len(self._references)))
                   for i in range(self._worker_processes)]

        for worker in workers:
            worker.start()

        worker_monitor = threading.Thread(target=monitor_workers, args=(workers, ))
        worker_monitor.start()

        for ref_id in self._references:
            task_queue.put(ref_id)

        # add poison pills to kill the workers
        for _ in range(self._worker_processes):
            task_queue.put(None)

        for worker in workers:
            worker.join()
        worker_monitor.join()

        logger.info("Merging tmp databases.")

        # Merge tmp databases
        tmp_dbs = [self._output + f".{i}"
                   for i in range(self._worker_processes)]
        self._db.merge_in_databases(tmp_dbs, merge_reads=True)

        logger.info("Creating database indices.")
        self._db.create_indices()


    def _get_references(self) -> set[str]:
        raise NotImplementedError("This method should be overriden by specific " + \
                                  "resquiggler preprocessor implementations.")


    def _get_resquiggled_data(
            self,
            ref_id: str,
            ref_seq: str
    ) -> tuple[Float[np.ndarray, "reads positions"],
               Float[np.ndarray, "reads positions"],
               list[str]]:
        raise NotImplementedError("This method should be overriden by specific " + \
                                  "resquiggler preprocessor implementations.")


    def _get_metadata(self) -> dict[str, Any]:
        raise NotImplementedError("This method should be overriden by specific " + \
                                  "resquiggler preprocessor implementations.")


    def _resquiggle(self,
                    idx,
                    task_queue,
                    lock,
                    current_transcript_id,
                    current_read_offset,
                    processed_transcripts,
                    num_transcripts):
        fasta_fh = Fasta(self._fasta_ref)

        tmp_db_out = PreprocessingDB(self._output + f".{idx}")
        tmp_db_out.setup(self._get_metadata())

        with closing(tmp_db_out.connect()) as conn:
            while True:
                ref_id = task_queue.get()
                if ref_id is None:
                    break

                ref_seq = str(fasta_fh[ref_id])
                intensity, dwell, reads = self._get_resquiggled_data(ref_id, ref_seq)
                read_invalid_ratios = get_reads_invalid_ratio(intensity.T)

                lock.acquire()
                transcript_id = current_transcript_id.value
                current_transcript_id.value += 1
                read_offset = current_read_offset.value
                current_read_offset.value += len(read_invalid_ratios)
                lock.release()

                reads_with_ratios = zip(reads, read_invalid_ratios)
                offsetted_reads = [(read_id, i + read_offset, invalid_ratio)
                                   for i, (read_id, invalid_ratio) in enumerate(reads_with_ratios)]
                signal_data_rows = self._process_rows_for_writing(
                        transcript_id,
                        [r[1] for r in offsetted_reads], # read internal ids
                        intensity.astype(np.float32),
                        dwell.astype(np.float32))

                with closing(conn.cursor()) as cursor:
                    cursor.execute("begin")
                    cursor.execute(INSERT_TRANSCRIPTS_QUERY, (transcript_id, ref_id))
                    cursor.executemany(INSERT_READS_QUERY, offsetted_reads)
                    cursor.execute("commit")
                PreprocessingDB.write_signal_data_rows(conn, signal_data_rows)

                lock.acquire()
                processed_transcripts.value += 1
                logger.info(f"Data for reference {processed_transcripts.value}/{num_transcripts} "
                            f"{ref_id} has been saved to the tmp database.")
                lock.release()


    def _process_rows_for_writing(self,
                                  transcript_id: int,
                                  read_ids: list[int],
                                  intensity: Float[np.ndarray, "positions"],
                                  dwell: Float[np.ndarray, "positions"]):
        for i in range(intensity.shape[0]):
            yield (transcript_id,
                   read_ids[i],
                   intensity[i].tobytes(),
                   dwell[i].tobytes())


class RemoraPreprocessor(Preprocessor):
    """
    Preprocess data using the Remora resquiggler.
    Remora resquiggling is done during the
    preprocessing (as oposed to Uncalled4 and
    Eventalign where the preprocessor will only
    populate the DB with signal alignment data
    that have already been produced by the resquiggler).
    The signal data would be transfered
    to an SQLite DB to be used by Nanocompore
    for later analysis.
    """
    def __init__(self,
                 fasta_ref: str,
                 pod5: str,
                 bam: str,
                 output: str,
                 kit: Kit,
                 max_reads: int,
                 threads: int):
        """
        Initialize a Remora preprocessor that will
        resquiggle all reads in the provided bam/pod5 pair.

        Parameters
        ----------
        fasta_ref : str
            Path to the fasta reference containing the transcripts.
        pod5 : str
            Path to the pod5 file containing the raw signal data.
        bam : str
            Path to the aligned bam file with the reads to resquiggle.
        output : str
            Path where the output database would be written to.
        kit : Kit
            The sequencing kit used.
        max_reads : int
            Maximum number of reads to resquiggle for each transcript.
        threads : int
            Number of parallel processes to use.
        """
        self._fasta_ref = fasta_ref
        self._pod5 = pod5
        self._bam = bam
        self._kit = kit
        self._max_reads = max_reads
        super().__init__(fasta_ref, output, threads)


    def _get_references(self) -> set[str]:
        return get_references_from_bam(self._bam)


    def _get_resquiggled_data(
            self,
            ref_id,
            ref_seq
    ) -> tuple[Float[np.ndarray, "reads positions"],
               Float[np.ndarray, "reads positions"],
               list[str]]:
        remora = Remora(self._pod5, self._bam, self._kit, self._max_reads)
        return remora.get_resquiggled_data(ref_id, ref_seq)


    def _get_metadata(self) -> dict[str, Any]:
        return {'fasta_ref': self._fasta_ref,
                'pod5': self._pod5,
                'bam': self._bam,
                'kit': self._kit,
                'max_reads': self._max_reads,
                'resquiggler': 'remora'}




import copy

import numpy as np
import pysam

from pyfaidx import Fasta

from nanocompore.config import Config
from nanocompore.uncalled4 import Uncalled4

from tests.common import BASIC_CONFIG

def test_get_data():
    config_yaml = copy.deepcopy(BASIC_CONFIG)
    # Note we use the same bam that contains
    # a single read so it doesn't matter
    # which read in the result we use for
    # the assertions.
    config_yaml['data'] = {
        'cond1': {
            'sample1': {
                'bam': 'tests/fixtures/uncalled4_sample.bam'
            }
        },
        'cond2': { 
            'sample2': {
                'bam': 'tests/fixtures/uncalled4_sample.bam'
            }
        }
    }
    config_yaml['depleted_condition'] = 'cond1'
    config_yaml['kit'] = 'RNA004'
    config = Config(config_yaml)
    ref = 'ENST00000234590.10|ENSG00000074800.16|OTTHUMG00000001773.11|OTTHUMT00000497295.2|ENO1-201|ENO1|1781|protein_coding|'
    fasta_fh = Fasta('tests/fixtures/uncalled4_reference.fa')
    ref_seq = str(fasta_fh[ref])

    bams = {sample: pysam.AlignmentFile(sample_def['bam'], 'rb')
            for cond, cond_def in config.get_data().items()
            for sample, sample_def in cond_def.items()}
    uncalled4 = Uncalled4(ref, len(ref_seq), bams, config.get_kit())
    tensor, sample_ids, condition_ids = uncalled4.get_data()
    # 1781 positions on the transcript, 2 reads and 2 vars (intensity and dwell)
    assert tensor.shape == (1781, 2, 2)
    # The read has two signal segments aligned
    # to reference positions [6, 562) and [604, 1779).
    # Let's look at the alignment in the 3'->5'
    # direction:
    # The first reference position included in
    # the measurements is 1778. Uncalled4
    # assigns the measurement to the most
    # influential (central) position,
    # which for RNA004 is 5th of the 9-mer.
    # Hence, Uncalled4 will assign the first
    # value to reference position 1774. We want
    # to use the kmer start (in 5'->3') direction
    # so we would expect the last measurement
    # to be for position 1770.
    # That would be the first value in the uc
    # tag in the bam, because the uc tag
    # lists values in the order of sequencing
    # (3'->5' for RNA).
    assert tensor[1770, 0, 0] == 7992
    assert np.isnan(tensor[1771, 0, 0])

    # The last kmer in the first segment covers
    # positions [604, 612]. This means that
    # Uncalled4 will have the last value for
    # the segment assigned to 608, but we'll
    # consider this value to be for 604 because
    # we use the kmer start (in 5'->3' direction).
    assert tensor[604, 0, 0] == -5072
    # Uncalled4 has NaNs (-32768) for the remainng
    # 4 reference positions of the kmer.
    assert np.isnan(tensor[603, 0, 0])
    assert np.isnan(tensor[602, 0, 0])
    assert np.isnan(tensor[601, 0, 0])
    assert np.isnan(tensor[600, 0, 0])

    # Then we have 4 NaNs in the form of -32768
    # values for the first 4 positions of the
    # second segment (in 3'->5' direction), i.e.
    # positions [554, 557].
    assert np.isnan(tensor[557, 0, 0])
    assert np.isnan(tensor[556, 0, 0])
    assert np.isnan(tensor[555, 0, 0])
    assert np.isnan(tensor[554, 0, 0])

    assert tensor[553, 0, 0] == -10820

    # The most extreme 5' position reported
    # by Uncalled4 is 10, which for us will
    # be position 6.
    # This is the last value in the bam
    # (because they're listed in the sequencing
    # order, i.e 3'->5' for RNA).
    assert tensor[6, 0, 0] == 3802
    # Position 5 shouldn't have a value.
    assert np.isnan(tensor[5, 0, 0])


def test_get_data_no_reads():
    config_yaml = copy.deepcopy(BASIC_CONFIG)
    # Note we use the same bam that contains
    # no reads for both conditions
    config_yaml['data'] = {
        'cond1': {
            'sample1': {
                'bam': 'tests/fixtures/empty.bam'
            }
        },
        'cond2': { 
            'sample2': {
                'bam': 'tests/fixtures/empty.bam'
            }
        }
    }
    config_yaml['depleted_condition'] = 'cond1'
    config_yaml['kit'] = 'RNA004'
    config = Config(config_yaml)
    ref = 'ENST00000234590.10|ENSG00000074800.16|OTTHUMG00000001773.11|OTTHUMT00000497295.2|ENO1-201|ENO1|1781|protein_coding|'
    fasta_fh = Fasta('tests/fixtures/uncalled4_reference.fa')
    ref_seq = str(fasta_fh[ref])

    bams = {sample: pysam.AlignmentFile(sample_def['bam'], 'rb')
            for cond, cond_def in config.get_data().items()
            for sample, sample_def in cond_def.items()}
    uncalled4 = Uncalled4(ref, len(ref_seq), bams, config.get_kit())
    tensor, sample_ids, condition_ids = uncalled4.get_data()

    assert tensor.shape == (1781, 0, 2)
    assert len(sample_ids) == 0
    assert len(condition_ids) == 0

